# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestEntryCommands(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = xmldata = "test_entry_commands.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("entry")

    def tearDown(self):
        support.root_withdraw()

    def test_validate_command(self):
        valid_values = ("Valid value1", "Valid value2")

        def entry_validate(action, newvalue):
            valid = False
            if action == "1":  # 1: insert 0: delete
                if newvalue in valid_values:
                    valid = True
            else:
                valid = True
            return valid

        def entry_invalid(newvalue):
            pass

        callback = {
            "entry_validate": entry_validate,
            "entry_invalid": entry_invalid,
        }

        self.builder.connect_callbacks(callback)

        self.widget.delete("0", tk.END)
        self.assertEqual("", self.widget.get())

        self.widget.insert("0", valid_values[0])
        self.assertEqual(valid_values[0], self.widget.get())

        self.widget.delete("0", tk.END)
        self.widget.insert("0", "Invalid value")
        self.assertEqual("", self.widget.get())

        self.widget.destroy()

    def test_invalid_command(self):
        invalid_text = []

        def entry_validate(action, newvalue):
            valid = False
            if action == "1":  # 1: insert 0: delete
                if newvalue == "Allowed":
                    valid = True
            else:
                valid = True
            return valid

        def entry_invalid(newvalue):
            invalid_text.append(newvalue)

        callback = {
            "entry_validate": entry_validate,
            "entry_invalid": entry_invalid,
        }

        self.widget.delete("0", tk.END)

        self.builder.connect_callbacks(callback)

        self.widget.insert("0", "Not Allowed")
        self.assertEqual(invalid_text[0], "Not Allowed")

        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
import tkinter as tk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from pygubu.widgets.floodgauge import Floodgauge
from pygubu.plugins.ttk.ttkstdwidgets import TTKProgressbar
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid

#
# Builder definition section
#


class FloodgaugeBO(TTKProgressbar):
    class_ = Floodgauge
    properties = TTKProgressbar.properties + ("mask", "text", "textvariable")


_builder_uid = f"{_plugin_uid}.Floodgauge"
register_widget(
    _builder_uid,
    FloodgaugeBO,
    "Floodgauge",
    ("ttk", _tab_widgets_label),
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


class TestToplevelMenuHelper(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_toplevelmenuhelper.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("toplevel")
        self.menuhelper = builder.get_object("topmenuhelper")
        self.menu = builder.get_object("topmenu")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.menu, tk.Menu)
        self.widget.destroy()

    def test_class_topmenu(self):
        menu1 = self.widget.nametowidget(self.widget.cget("menu"))
        self.assertEqual(menu1, self.menu)
        self.widget.destroy()



from .registry import StockRegistry
from .loader import StockImageCache
from .exceptions import StockImageException
from .config import TK_BITMAP_FORMATS, TK_IMAGE_FORMATS, TK_PHOTO_FORMATS

DefaultRegistry = StockRegistry()

StockImage = StockImageCache(None, DefaultRegistry)



import logging
import platform
import tkinter as tk

from typing import List
from pygubu.utils.widget import iter_to_toplevel


logger = logging.getLogger(__name__)


class MouseWheelCommand:
    def __init__(self, view_command, factor=1):
        self.view_command = view_command
        self.factor = factor


class UnixMouseWheelCommandTk9(MouseWheelCommand):
    def __call__(self, event=None) -> bool:
        can_keep_scrolling = False
        self.view_command(
            "scroll",
            (-1) * int((event.delta / 120) * self.factor),
            "units",
        )
        scroll_rs = self.view_command()
        if scroll_rs is not None:
            can_keep_scrolling = scroll_rs[0] != 0.0
            if event.delta < 0:
                can_keep_scrolling = scroll_rs[1] != 1.0
        return can_keep_scrolling


class UnixMouseWheelCommandTk8(MouseWheelCommand):
    def __call__(self, event=None) -> bool:
        can_keep_scrolling = False
        if event.num == 4:
            self.view_command("scroll", (-1) * self.factor, "units")
            scroll_rs = self.view_command()
            if scroll_rs is not None:
                can_keep_scrolling = scroll_rs[0] != 0.0
        elif event.num == 5:
            self.view_command("scroll", self.factor, "units")
            scroll_rs = self.view_command()
            if scroll_rs is not None:
                can_keep_scrolling = scroll_rs[1] != 1.0
        return can_keep_scrolling


class WindowsMouseWheelCommand(MouseWheelCommand):
    def __call__(self, event=None) -> bool:
        can_keep_scrolling = False
        self.view_command(
            "scroll", (-1) * int((event.delta / 120) * self.factor), "units"
        )
        scroll_rs = self.view_command()
        if scroll_rs is not None:
            can_keep_scrolling = scroll_rs[0] != 0.0
            if event.delta < 0:
                can_keep_scrolling = scroll_rs[1] != 1.0
        return can_keep_scrolling


class DarwingMouseWheelCommand(MouseWheelCommand):
    def __call__(self, event=None) -> bool:
        can_keep_scrolling = False
        self.view_command("scroll", event.delta, "units")
        scroll_rs = self.view_command()
        if scroll_rs is not None:
            can_keep_scrolling = scroll_rs[0] != 0.0
            if event.delta < 0:
                can_keep_scrolling = scroll_rs[1] != 1.0
        return can_keep_scrolling


class UnknownMouseWheelCommand(MouseWheelCommand):
    def __call__(self, event=None) -> bool:
        # Unknown platform scroll method
        return False


class AppBindManagerBase(object):
    CLASS_NAME = "TAppBindManager"
    FLAG_NAME = "_appbind_visited"
    # Acces to tk instances
    masters: List[tk.Widget] = []
    # Mouse wheel support
    mw_listeners = []  # Mousewheel listeners

    @classmethod
    def master_for_path(cls, widget_path: str):
        master_found = None
        for master in cls.masters:
            exists = master.tk.eval(f"winfo exists {widget_path}")
            if exists:
                master_found = master
                break
        return master_found

    @classmethod
    def on_mousewheel_enter(cls, event):
        """Prepare widget for future mousewheel event."""
        widget_below = event.widget
        if not isinstance(widget_below, tk.Widget):
            return
        if not hasattr(widget_below, cls.FLAG_NAME):
            flag_value = cls.has_mousewheel_listener(widget_below)
            setattr(widget_below, cls.FLAG_NAME, flag_value)
            if flag_value:
                # print(f"Preparing widget for mousewheel {widget_below}")
                # cname = widget_below.winfo_class()
                tags = list(widget_below.bindtags())
                tags.insert(0, cls.CLASS_NAME)
                widget_below.bindtags(tags)

    @classmethod
    def has_mousewheel_listener(cls, widget: tk.Widget) -> bool:
        for w in iter_to_toplevel(widget):
            if w in cls.mw_listeners:
                return True
        return False

    @classmethod
    def on_mousewheel(cls, event):
        """
        Manage Application level mousewheel event

        Here we expect that the widget in wm_listeners list,
        have a method "on_mousewheel" created with the
        make_onmousewheel_cb function of this class.
        """
        # print("on_mousewheel, cls", event)
        widget_below = event.widget
        if not isinstance(widget_below, tk.Widget):
            try:
                master = cls.master_for_path(widget_below)
                widget_below = master.winfo_containing(
                    event.x_root, event.y_root
                )
            except (AttributeError, KeyError):
                widget_below = None
        # scroll_performed = False
        if widget_below:
            for w in iter_to_toplevel(widget_below):
                if w in cls.mw_listeners:
                    can_keep_scrolling = w.on_mousewheel(event)
                    if can_keep_scrolling:
                        # scroll_performed = True
                        break
        # if scroll_performed:
        #    print("scroll performed")

        # For now just stop event propagation.
        return "break"

    @classmethod
    def mousewheel_bind(cls, widget):
        if widget not in cls.mw_listeners:
            cls.mw_listeners.append(widget)

    @classmethod
    def mousewheel_unbind(cls, widget):
        if widget in cls.mw_listeners:
            cls.mw_listeners.remove(widget)

    @classmethod
    def init_mousewheel_binding(cls, master):
        top = master.winfo_toplevel()
        if top not in cls.masters:
            cls.masters.append(top)
            # Bind to Enter event to prepare widgets
            top.bind_all("<Enter>", cls.on_mousewheel_enter, add="+")

            # Bind Wheel events to a specific class name
            # so we cant "break" the event and stop propagation
            _os = platform.system()
            if _os in ("Linux", "OpenBSD", "FreeBSD"):
                if tk.TkVersion >= 9:
                    top.bind_class(
                        cls.CLASS_NAME,
                        "<MouseWheel>",
                        cls.on_mousewheel,
                        add="+",
                    )
                else:
                    top.bind_class(
                        cls.CLASS_NAME, "<4>", cls.on_mousewheel, add="+"
                    )
                    top.bind_class(
                        cls.CLASS_NAME, "<5>", cls.on_mousewheel, add="+"
                    )
            else:
                # Windows and MacOS
                top.bind_class(
                    cls.CLASS_NAME,
                    "<MouseWheel>",
                    cls.on_mousewheel,
                    add="+",
                )

    @classmethod
    def make_onmousewheel_cb(cls, widget, orient, factor=1):
        """Create a callback to manage mousewheel events

        orient: string (posible values: ('x', 'y'))
        widget: widget that implement tk xview and yview methods
        """
        _os = platform.system()
        view_command = getattr(widget, orient + "view")
        on_mousewheel = None
        if _os in ("Linux", "OpenBSD", "FreeBSD"):
            if tk.TkVersion >= 9:
                on_mousewheel = UnixMouseWheelCommandTk9(view_command, factor)
            else:
                on_mousewheel = UnixMouseWheelCommandTk8(view_command, factor)
        elif _os == "Windows":
            on_mousewheel = WindowsMouseWheelCommand(view_command, factor)
        elif _os == "Darwin":
            on_mousewheel = DarwingMouseWheelCommand(view_command, factor)
        else:
            on_mousewheel = UnknownMouseWheelCommand(view_command, factor)

        return on_mousewheel


class ApplicationLevelBindManager(AppBindManagerBase):
    ...



# encoding: utf-8
__all__ = ["ToolTip"]

import tkinter as tk
import tkinter.ttk as ttk
from contextlib import suppress


class TooltipBase:
    """Base class for tooltips using a Label widget."""

    tipwindows = {}
    label_class = tk.Label

    def __init__(self, widget: tk.Widget, /, **label_options):
        self.widget = widget
        self.x: int = 0
        self.y: int = 0
        self.label_options: dict = label_options
        self._bind_ids = (
            self.widget.bind("<Enter>", self.on_enter, add=True),
            self.widget.bind("<Leave>", self.on_leave, add=True),
        )
        # save reference for simple acces from wiget
        widget.stooltip = self

    # Add text property to mantain backward comatibility?
    @property
    def text(self):
        return self.label_options.get("text", "")

    @text.setter
    def text(self, value):
        self.label_options["text"] = value

    def inside_wbbox(self, rx, ry):
        bbox = self._calc_bbox(self.widget, True)
        inside = False
        if (bbox[0] <= rx <= bbox[2]) and (bbox[1] <= ry <= bbox[3]):
            inside = True
        return inside

    def _calc_bbox(self, widget: tk.Widget, screen=False):
        rx = widget.winfo_x()
        ry = widget.winfo_y()
        if screen:
            rx = widget.winfo_rootx()
            ry = widget.winfo_rooty()
        x2 = rx + widget.winfo_width()
        y2 = ry + widget.winfo_height()
        return (rx, ry, x2, y2)

    def _calc_final_pos(self, ttwidth, ttheight):
        rx, ry, rcx, rcy = self._calc_bbox(self.widget, True)
        w = rcx - rx
        h = rcy - ry
        sh = self.widget.winfo_screenheight()
        sw = self.widget.winfo_screenwidth()
        x = y = 0
        # final_region = None
        regions = (
            "se-al",
            "se-ar",
            "sw-al",
            "sw-ar",
            "nw-al",
            "nw-ar",
            "ne-al",
            "ne-ar",
        )
        for region in regions:
            if region == "ne-al":
                x = rx
                y = ry - ttheight
            elif region == "ne-ar":
                x = rx - ttwidth
                y = ry - ttheight
            elif region == "nw-al":
                x = rx + w
                y = ry - ttheight
            elif region == "nw-ar":
                x = rx + w - ttwidth
                y = ry - ttheight
            elif region == "se-al":
                x = rx
                y = ry + h
            elif region == "se-ar":
                x = rx - ttwidth
                y = ry + h
            elif region == "sw-al":
                x = rx + w
                y = ry + h
            elif region == "sw-ar":
                x = rx + w - ttwidth
                y = ry + h
            x2 = x + ttwidth
            y2 = y + ttheight
            if (x > 0 and x2 < sw) and (y > 0 and y2 < sh):
                # final_region = region
                break
        # print(final_region, x, x2, y, y2)
        return (x, y)

    def on_enter(self, event):
        self.showtip()

    def on_leave(self, event):
        self.hidetip()

    def __del__(self):
        with suppress(tk.TclError):
            if self.widget.winfo_exists():
                self.widget.unbind("<Enter>", self._bind_ids[0])
                self.widget.unbind("<Leave>", self._bind_ids[1])

    def _tipwindow_for(self, widget: tk.Widget) -> tk.Toplevel:
        root = widget.winfo_toplevel()
        root_id = id(root)
        tipwindow = self.tipwindows.get(root_id, None)
        if tipwindow is None:
            tipwindow = tk.Toplevel(root)
            tipwindow.wm_overrideredirect(1)
            with suppress(tk.TclError):
                # For Mac OS
                tipwindow.tk.call(
                    "::tk::unsupported::MacWindowStyle",
                    "style",
                    tipwindow._w,
                    "help",
                    "noActivates",
                )
            tipwindow.label = self._label_create(tipwindow)
            self.tipwindows[root_id] = tipwindow
        return tipwindow

    def _label_create(self, tipwindow: tk.Toplevel) -> tk.Widget:
        label = self.label_class(tipwindow)
        label.pack(expand=True, fill=tk.BOTH)
        return label

    def _label_config(self, label):
        label.configure(**self.label_options)

    def showtip(self):
        """Display text in tooltip window"""
        tipwindow = self._tipwindow_for(self.widget)
        label = tipwindow.label
        self._label_config(label)
        x, y = self._calc_final_pos(
            label.winfo_reqwidth(), label.winfo_reqheight()
        )
        tipwindow.wm_geometry(f"+{x}+{y}")
        tipwindow.deiconify()

    def hidetip(self):
        tipwindow = self._tipwindow_for(self.widget)
        tipwindow.withdraw()

    def configure(self, **kw):
        """Updates label options.
        Similar to configure method of tk widgets.
        """
        self.label_options.update(kw)


class Tooltip(TooltipBase):
    """A simple tooltip class based on tk.Label"""

    default_config = dict(
        text="undefined tooltip",
        font=("tahoma", "9", "normal"),
        background="#ffffe0",
        foreground="black",
        justify=tk.LEFT,
        wraplength="300p",
        relief=tk.SOLID,
        borderwidth="1p",
        padx="2p",
        pady="2p",
    )

    def __init__(self, widget: tk.Widget, /, **label_options):
        options = self.default_config.copy()
        options.update(label_options)
        super().__init__(widget, **options)


# Maintain old class name for now
ToolTip = Tooltip


class Tooltipttk(TooltipBase):
    """A simple tooltip class based on ttk.Label"""

    STYLE_NAME = "Tooltipttk.TLabel"
    STYLE_OPTIONS = dict()
    tipwindows = {}
    label_class = ttk.Label
    default_config = dict(style=STYLE_NAME)
    clear_config = dict(
        text="",
        font="",
        foreground="",
        background="",
        justify="left",
        wraplength="",
        relief="",
        borderwidth="",
        style=STYLE_NAME,
    )

    def __init__(self, widget: tk.Widget, /, **label_options):
        options = self.default_config.copy()
        options.update(label_options)
        super().__init__(widget, **options)

    def _label_config(self, label):
        label.configure(**self.clear_config)
        label.configure(**self.label_options)


def create(widget, text, **label_options):
    """Creates a Tooltip using a tk.Label."""
    label_options["text"] = text
    tooltip = Tooltip(widget, **label_options)
    return tooltip


def create_ttk(widget, text, **label_options):
    """Creates a Tooltip using a ttk.Label."""
    label_options["text"] = text
    tooltip = Tooltipttk(widget, **label_options)
    return tooltip


if __name__ == "__main__":

    def build_test(root, tk_options: dict = None, ttk_options: dict = None):
        tk_options = {} if tk_options is None else tk_options
        ttk_options = {} if ttk_options is None else ttk_options

        for idx in range(0, 2):
            b = tk.Button(root, text=f"Button({idx})")
            b.pack(pady="2p")
            text = f"A tooltip for button {idx}!!"
            create(b, text, **tk_options)

        for idx in range(2, 4):
            b = ttk.Button(root, text=f"Button({idx})")
            b.pack(pady="2p")
            text = f"A ttk tooltip for button {idx}!!"
            create_ttk(b, text, **ttk_options)

    root = tk.Tk()
    root.title("Root 1")
    root.minsize(300, 200)
    tk_options = {}  # dict(background="darkgreen", foreground="white")
    ttk_options = {}
    build_test(root)

    root2 = tk.Tk()
    root2.title("Root 2")
    root2.minsize(300, 200)

    s = ttk.Style(root2)
    s.configure(Tooltipttk.STYLE_NAME, foreground="white", background="red")
    ttk_options = dict()

    tk_options = dict(background="blue", foreground="white")
    build_test(root2, tk_options)

    root.mainloop()



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support
from pygubu.widgets.accordionframe import AccordionFrame


class TestAccordionFrame(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_accordionframe.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainframe")
        self.canvas = builder.get_object("accordionframe1")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.canvas, AccordionFrame)
        self.widget.destroy()



# encoding: utf-8
import os
import sys
import unittest
import datetime
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support
from pygubu.widgets.calendarframe import CalendarFrame


class TestCalendarFrame(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_calendarframe.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.mainwindow = builder.get_object("mainwindow")
        self.widget = builder.get_object("calendar")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, CalendarFrame)
        self.widget.destroy()

    def test_selection(self):
        self.assertEqual(None, self.widget.selection)

        testdate = datetime.datetime(1980, 9, 15)
        self.widget.select_day(15, 9, 1980)

        self.assertEqual(testdate, self.widget.selection)
        self.widget.destroy()

    def test_select_day(self):
        testdate = datetime.datetime(1980, 9, 15)
        self.widget.select_day(15, 9, 1980)
        self.assertEqual(testdate, self.widget.selection)
        self.widget.destroy()

    def test_event_CalendarDateSelected(self):
        success = []

        class AnObject:
            def on_date_selected(self, event=None):
                success.append(1)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)
        self.widget.select_day(15, 9, 1980)
        self.widget.update()

        self.assertTrue(success)
        self.widget.destroy()

    def test_uiset_options(self):
        widget = self.widget
        self.assertEqual("#f0e3f0", widget["calendarbg"])
        self.assertEqual("#0000a1", widget["calendarfg"])

        self.assertEqual("#abb8f0", widget["headerbg"])
        self.assertEqual("#ffffff", widget["headerfg"])

        self.assertEqual("#f06ba4", widget["markbg"])
        self.assertEqual("#ffffff", widget["markfg"])

        self.assertEqual("#d926ff", widget["selectbg"])
        self.assertEqual("#ffff00", widget["selectfg"])

        self.assertEqual("9", str(widget["month"]))
        self.assertEqual("2000", str(widget["year"]))

        self.widget.destroy()



# encoding: utf-8
import tkinter as tk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.widgets.accordionframe import AccordionFrame
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class AccordionFrameBO(TTKFrame):
    class_ = AccordionFrame
    allowed_children = (f"{_plugin_uid}.AccordionFrameGroup",)
    _img_properties = ("img_expand", "img_collapse")
    properties = TTKFrame.properties + _img_properties
    tkimage_properties = TTKFrame.tkimage_properties + _img_properties


_builder_uid = f"{_plugin_uid}.AccordionFrame"

register_widget(
    _builder_uid,
    AccordionFrameBO,
    "AccordionFrame",
    (_tab_widgets_label, "ttk"),
)


class AccordionFrameGroupBO(BuilderObject):
    allowed_parents = (f"{_plugin_uid}.AccordionFrame",)
    properties = ("label", "expanded", "compound", "style")
    layout_required = False
    container = True
    container_layout = True
    allow_bindings = False

    def realize(self, parent, extra_init_args: dict = None):
        args = self._get_init_args(extra_init_args)
        master = parent.get_child_master()
        gid = self.wmeta.identifier
        self.widget = master.add_group(gid, **args)
        self.widget._accordion = parent.widget
        return self.widget

    def _process_property_value(self, pname, value):
        if pname == "expanded":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _set_property(self, target_widget, pname, value):
        if pname in self.properties:
            propvalue = self._process_property_value(pname, value)
            self.widget._accordion.group_config(
                self.wmeta.identifier, **{pname: propvalue}
            )
        else:
            super()._set_property(target_widget, pname, value)


_builder_uid = f"{_plugin_uid}.AccordionFrameGroup"

register_widget(
    _builder_uid,
    AccordionFrameGroupBO,
    "AccordionFrame.Group",
    (_tab_widgets_label, "ttk"),
)



import tkinter as tk
import tkinter.ttk as ttk
from pygubu.utils.widget import WidgetConfigureMixin


"""Floodgauge Widget

Based on ttkbootstrap Floodgauge.
"""


class FloodgaugeStyleManager:
    """Manage PygubuFloodgauge style

    Tasks:
        - register style if not defined
        - update style if theme changed
        - do nothing if style is previously defined
          (if using pygubu theming).
    """

    TKCLASS_NAME: str = "Floodgauge"
    STYLE_NAME_H: str = None
    STYLE_NAME_V: str = None
    STYLE_LAYOUT_H: list = None
    STYLE_LAYOUT_V: list = None
    STYLE_CONF: dict = None
    STYLE_INITIALIZED = False
    STYLE_MANAGED_EXTERNALLY = False
    CONFIGURED_THEMES = None

    @classmethod
    def define_layout(cls, master):
        if cls.STYLE_INITIALIZED or cls.STYLE_MANAGED_EXTERNALLY:
            return

        H_STYLE = f"Horizontal.T{cls.TKCLASS_NAME}"
        V_STYLE = f"Vertical.T{cls.TKCLASS_NAME}"

        cls.STYLE_NAME_H = H_STYLE
        cls.STYLE_NAME_V = V_STYLE

        style = ttk.Style(master)
        try:
            h_layout = style.layout(H_STYLE)
            v_layout = style.layout(V_STYLE)
            if h_layout and v_layout:
                cls.STYLE_MANAGED_EXTERNALLY = True
                cls.STYLE_LAYOUT_H = h_layout
                cls.STYLE_LAYOUT_V = v_layout
                return
        except tk.TclError:
            pass

        h_element = H_STYLE.replace(".TF", ".F")
        v_element = V_STYLE.replace(".TF", ".F")
        style.element_create(f"{h_element}.trough", "from", "clam")
        style.element_create(f"{h_element}.pbar", "from", "default")
        style.element_create(f"{v_element}.trough", "from", "clam")
        style.element_create(f"{v_element}.pbar", "from", "default")

        h_layout = None
        v_layout = None
        if tk.TkVersion >= 9:
            h_layout, v_layout = cls.floodgauge_layout_tk9(h_element, v_element)
        else:
            h_layout, v_layout = cls.floodgauge_layout_tk8(h_element, v_element)

        bg_color = style.lookup("TProgressbar", "background")
        default_conf = dict(
            background=bg_color,
            borderwidth=1,
            font="-size 14",
            thickness=50,
            pbarrelief=tk.FLAT,
            justify=tk.CENTER,
            anchor=tk.CENTER,
        )
        style.layout(H_STYLE, h_layout)
        style.configure(H_STYLE, **default_conf)
        style.layout(V_STYLE, v_layout)
        style.configure(V_STYLE, **default_conf)

        cls.CONFIGURED_THEMES = [style.theme_use()]
        cls.STYLE_LAYOUT_H = h_layout
        cls.STYLE_LAYOUT_V = v_layout
        cls.STYLE_CONF = default_conf
        cls.STYLE_INITIALIZED = True

    @classmethod
    def floodgauge_layout_tk9(cls, h_element: str, v_element: str) -> tuple:
        h_layout = [
            (
                f"{h_element}.trough",
                {
                    "children": [
                        (f"{h_element}.pbar", {"side": "left", "sticky": "ns"}),
                        (
                            f"{h_element}.ctext",
                            # {"side": "left", "sticky": ""}
                            {"side": "left", "sticky": "", "expand": True},
                        ),
                    ],
                    "sticky": "nswe",
                },
            )
        ]

        v_layout = [
            (
                f"{v_element}.trough",
                {
                    "children": [
                        (
                            f"{v_element}.pbar",
                            {"side": "bottom", "sticky": "we"},
                        ),
                        (f"{v_element}.ctext", {"sticky": "", "expand": True}),
                    ],
                    "sticky": "nswe",
                },
            )
        ]
        return h_layout, v_layout

    @classmethod
    def floodgauge_layout_tk8(cls, h_element: str, v_element: str) -> tuple:
        h_layout = [
            (
                f"{h_element}.trough",
                {
                    "children": [
                        (
                            f"{h_element}.pbar",
                            {"sticky": "ns"},
                        ),
                        (f"{cls.TKCLASS_NAME}.label", {"sticky": ""}),
                    ],
                    "sticky": "nsew",
                },
            )
        ]
        v_layout = [
            (
                f"{v_element}.trough",
                {
                    "children": [
                        (
                            f"{v_element}.pbar",
                            {"sticky": "ew"},
                        ),
                        (f"{cls.TKCLASS_NAME}.label", {"sticky": ""}),
                    ],
                    "sticky": "nsew",
                },
            )
        ]
        return h_layout, v_layout

    @classmethod
    def reconfigure_layout(cls, master):
        if cls.STYLE_MANAGED_EXTERNALLY:
            return
        theme = master.tk.eval("return $ttk::currentTheme")
        if theme not in cls.CONFIGURED_THEMES:
            s = ttk.Style(master)
            bg_color = s.lookup("TProgressbar", "background")
            conf = cls.STYLE_CONF
            conf.update(background=bg_color)
            s.configure(cls.STYLE_NAME_H, **conf)
            s.configure(cls.STYLE_NAME_V, **conf)
            cls.CONFIGURED_THEMES.append(theme)

    @classmethod
    def style_for_orient(cls, orient):
        return cls.STYLE_NAME_H if orient == "horizontal" else cls.STYLE_NAME_V

    @classmethod
    def layout_for_orient(cls, orient):
        return (
            cls.STYLE_LAYOUT_H if orient == "horizontal" else cls.STYLE_LAYOUT_V
        )


class FloodgaugeBase(WidgetConfigureMixin, ttk.Progressbar):
    """A widget that shows the status of a long-running operation
    with an optional text indicator.

    Similar to the `ttk.Progressbar`, this widget can operate in
    two modes. *determinate* mode shows the amount completed
    relative to the total amount of work to be done, and
    *indeterminate* mode provides an animated display to let the
    user know that something is happening.
    """

    FGSM = FloodgaugeStyleManager()

    def __init__(
        self,
        master=None,
        *,
        mask=None,
        variable=None,
        textvariable=None,
        value=0,
        text="",
        style=None,
        class_=None,
        **kw,
    ):
        self.FGSM.define_layout(master)
        self._traceid = None
        self._pmask = mask
        self._ptextvar = textvariable
        self._pvariable = variable
        if variable is None:
            self._pvariable = tk.IntVar(master, value=value)
        if textvariable is None:
            self._ptextvar = tk.StringVar(master, value=text)

        if style is None:
            orient = kw.get("orient", "horizontal")
            style = self.FGSM.style_for_orient(orient)
        if class_ is None:
            class_ = self.FGSM.TKCLASS_NAME

        kw["value"] = value
        kw["variable"] = self._pvariable
        kw["style"] = style
        kw["class_"] = class_
        super().__init__(master, **kw)
        if self._pmask is not None:
            self._set_mask()
        self.bind("<<ThemeChanged>>", self._on_theme_change)
        self.bind("<<Configure>>", self._on_theme_change)

    def _widget_cget(self, option):
        if option == "value":
            return self._pvariable.get()
        if option == "text":
            return self._ptextvar.get()
        if option == "textvariable":
            return self._ptextvar
        if option == "mask":
            return self._pmask
        return super()._widget_cget(option)

    def _configure_get(self, option):
        if option in ("value", "text", "mask", "textvariable"):
            return self._widget_cget(option)
        return super()._configure_get(option)

    def _configure_set(self, **kw):
        update_text = False
        if "variable" in kw:
            new_var = kw["variable"]
            self._unset_mask()
            self._pvariable = new_var
            if new_var and self._pmask:
                self._set_mask()
            update_text = True
        if "value" in kw:
            self._pvariable.set(kw.pop("value"))
            update_text = True
        if "text" in kw:
            self._ptextvar.set(kw.pop("text"))
            update_text = True
        if "mask" in kw:
            self._pmask = kw.pop("mask")
            if self._pmask:
                self._set_mask()
            update_text = True
        if "orient" in kw:
            new_orient = kw.get("orient")
            current = super()._configure_get("orient")
            if new_orient != current:
                style = kw.get("style", super()._configure_get("style"))
                if new_orient[1:] not in style:
                    kw["style"] = self.FGSM.style_for_orient(new_orient)
        if update_text:
            self._set_widget_text()
        return super()._configure_set(**kw)

    def _set_mask(self):
        if self._traceid is None:
            self._traceid = self._pvariable.trace_add(
                "write", self._set_widget_text
            )

    def _unset_mask(self):
        if self._traceid is not None:
            self._pvariable.trace_remove("write", self._traceid)
        self._traceid = None

    def _set_widget_text(self, *_):
        if self._pmask is None:
            text = self._ptextvar.get()
        else:
            value = self._pvariable.get()
            text = self._pmask.format(value)
        super()._configure_set(text=text)

    def _on_theme_change(self, event: tk.Event = None):
        self.FGSM.reconfigure_layout(self)


class FloodgaugeTk8(FloodgaugeBase):
    INSTANCE_COUNT = 0

    def __init__(self, master=None, **kw):
        self.FGSM.define_layout(master)

        self._uid = self._new_uid()
        self._pfont = kw.pop("font", "helvetica 10")

        orient = kw.get("orient", "horizontal")
        default_style = self.FGSM.style_for_orient(orient)
        user_defined_style = kw.get(
            "style", default_style
        )  # user defined style

        kw["style"] = self._new_instance_style(
            master, orient, user_defined_style
        )
        super().__init__(master, **kw)
        self._set_widget_text()

    @classmethod
    def _new_uid(cls):
        cls.INSTANCE_COUNT += 1
        return f"fg{cls.INSTANCE_COUNT}"

    def _set_widget_text(self, *_):
        ttkstyle = self.cget("style")
        if self._pmask is None:
            text = self._ptextvar.get()
        else:
            value = self._pvariable.get()
            text = self._pmask.format(value)
        self.tk.call("ttk::style", "configure", ttkstyle, "-text", text)
        self.tk.call("ttk::style", "configure", ttkstyle, "-font", self._pfont)

    def _new_instance_style(self, master, orient: str, style: str) -> str:
        instance_style = style
        uid = f"{self._uid}."

        if not style.startswith(uid):
            instance_style = f"{uid}{style}"

        layout_exists = False
        s = ttk.Style(master)
        try:
            s.layout(instance_style)
        except tk.TclError:
            pass
        if not layout_exists:
            s.layout(instance_style, self.FGSM.layout_for_orient(orient))
            # print(f"creating layout: {instance_style}")
        return instance_style

    def _on_theme_change(self, event: tk.Event = None):
        super()._on_theme_change(event)
        self._set_widget_text()

    def _widget_cget(self, option):
        if option == "font":
            return self._pfont
        return super()._widget_cget(option)

    def _configure_get(self, option):
        if option == "font":
            return self._widget_cget(option)
        return super()._configure_get(option)

    def _configure_set(self, **kw):
        update_text = False
        update_style = False
        if "orient" in kw or "style" in kw:
            update_style = True
        if update_style:
            orient = kw.get("orient", self.cget("orient"))
            new_style = kw.pop("style", self.FGSM.style_for_orient(orient))
            style = self._new_instance_style(self, orient, new_style)
            # super()._configure_set(style=style)
            kw["style"] = style
            update_text = True
        if "variable" in kw or "value" in kw or "text" in kw or "mask" in kw:
            update_text = True
        if "font" in kw:
            self._pfont = kw.pop("font")
            update_text = True
        if update_text:
            self._set_widget_text()
        return super()._configure_set(**kw)


if tk.TkVersion >= 9:

    class Floodgauge(FloodgaugeBase):
        ...

else:
    Floodgauge = FloodgaugeTk8


if __name__ == "__main__":
    root = tk.Tk()
    gauge = Floodgauge(root, mask="{}%")
    gauge.pack(expand=True, fill="both")
    options = ("value", "text", "mask", "font", "textvariable", "variable")
    for option in options:
        value = gauge.cget(option)
        print(f"The gauge {option} is:", value, type(value))

    scale = ttk.Scale(
        root, variable=gauge.cget("variable"), value=0, from_=0, to=100
    )
    scale.pack()

    root.mainloop()



import tkinter as tk
from pygubu.api.v1 import BuilderObject
from pygubu.i18n import _


_plugin_forms_uid = "pygubu.forms"
_tab_form_widgets_label = _("Pygubu Forms")


class WidgetBOMixin:
    """Manages base widget properties."""

    FIELD_NAME_PROP = "field_name"
    base_properties = (FIELD_NAME_PROP,)

    # @classmethod
    # def properties(cls):
    # if isinstance(super().properties, tuple):
    # return super().properties + cls.base_properties
    # return super().properties() + cls.base_properties

    # @classmethod
    # def ro_properties(cls):
    # if isinstance(super().ro_properties, tuple):
    # return super().ro_properties + cls.base_properties
    # return super().ro_properties() + cls.base_properties

    def _get_init_args(self, extra_init_args: dict = None):
        args = super()._get_init_args(extra_init_args)
        name = args.get(self.FIELD_NAME_PROP, None)
        if not name:
            args[self.FIELD_NAME_PROP] = self.wmeta.identifier
        return args

    def _code_get_init_args(self, code_identifier):
        args = super()._code_get_init_args(code_identifier)
        field_name_value = args.get(self.FIELD_NAME_PROP, None)
        if not field_name_value:
            field_name_value = self.wmeta.identifier
            args[self.FIELD_NAME_PROP] = self._code_process_property_value(
                code_identifier, self.FIELD_NAME_PROP, field_name_value
            )
        return args



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.utils.font import tkfontstr_to_tuple
from tkintertable import TableCanvas
from ..tkintertable import _designer_tab_label, _plugin_uid


class TableCanvasBuilder(BuilderObject):
    class_ = TableCanvas
    OPTIONS_CUSTOM = (
        "read_only",
        "width",
        "height",
        "bgcolor",
        "fgcolor",
        "rows",
        "cols",
        "cellwidth",
        "maxcellwidth",
        "rowheight",
        "horizlines",
        "vertlines",
        "alternaterows",
        "autoresizecols",
        "linewidth",
        "rowheaderwidth",
        "showkeynamesinheader",
        "thefont",
        "entrybackgr",
        "grid_color",
        "selectedcolor",
        "rowselectedcolor",
        "multipleselectioncolor",
    )
    ro_properties = OPTIONS_CUSTOM
    layout_required = False

    def _process_property_value(self, pname, value):
        if pname in (
            "rows",
            "cols",
            "cellwidth",
            "maxcellwidth",
            "rowheight",
            "rowheaderwidth",
        ):
            value = int(value)
        elif pname == "linewidth":
            value = float(value)
        elif pname in ("read_only", "showkeynamesinheader"):
            value = tk.getboolean(value)
        elif pname == "thefont":
            value = tkfontstr_to_tuple(value)
        elif pname in (
            "horizlines",
            "vertlines",
            "alternaterows",
            "autoresizecols",
        ):
            value = int(tk.getboolean(value))
        return value

    def layout(self, target=None, *, forget=False):
        self.widget.show()

    def code_layout(self, targetid=None, parentid=None):
        if targetid is None:
            targetid = self.code_identifier()
        return [f"{targetid}.show()"]

    def _code_process_property_value(self, targetid, pname, value: str):
        pvalue = self._process_property_value(pname, value)
        if pname in ("thefont", "showkeynamesinheader"):
            return pvalue
        pvalue = f"{pvalue}"
        return super()._code_process_property_value(targetid, pname, pvalue)


_builder_uid = f"{_plugin_uid}.TableCanvas"
register_widget(
    _builder_uid,
    TableCanvasBuilder,
    "TableCanvas",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "read_only",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "width",
    "dimensionentry",
)
register_custom_property(
    _builder_uid,
    "height",
    "dimensionentry",
)
register_custom_property(
    _builder_uid,
    "bgcolor",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "fgcolor",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "rows",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "cols",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "cellwidth",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "maxcellwidth",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "rowheight",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "horizlines",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "vertlines",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "alternaterows",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "autoresizecols",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "linewidth",
    "realnumber",
)
register_custom_property(
    _builder_uid,
    "rowheaderwidth",
    "naturalnumber",
)
register_custom_property(
    _builder_uid,
    "showkeynamesinheader",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "thefont",
    "fontentry",
)
register_custom_property(
    _builder_uid,
    "entrybackgr",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "grid_color",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "selectedcolor",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "rowselectedcolor",
    "colorentry",
)
register_custom_property(
    _builder_uid,
    "multipleselectioncolor",
    "colorentry",
)



import colorsys
import tkinter as tk


class ColorUtil:
    """Color utilities using tkinter functions."""

    def __init__(self, master):
        self.tk_master: tk.Widget = master

    # Color utilities begin
    def make_transparent(self, alpha, foreground, background="#ffffff"):
        """Simulate color transparency."""
        fg = tuple((c // 256 for c in self.tk_master.winfo_rgb(foreground)))
        bg = tuple((c // 256 for c in self.tk_master.winfo_rgb(background)))
        rgb_float = [alpha * c1 + (1 - alpha) * c2 for (c1, c2) in zip(fg, bg)]
        rgb_int = [int(x) for x in rgb_float]
        return "#{:02x}{:02x}{:02x}".format(*rgb_int)

    def hex_to_rgb(self, color: str):
        return tuple((c // 256 for c in self.tk_master.winfo_rgb(color)))

    def update_hsv(self, color, hd=0, sd=0, vd=0):
        """Modify the hue, saturation, and/or value of a given hex
        color value by specifying the _delta_.

        Parameters:

            color (str):
                A hexadecimal color value to adjust.

            hd (float):
                % change in hue, _hue delta_.

            sd (float):
                % change in saturation, _saturation delta_.

            vd (float):
                % change in value, _value delta_.

        Returns:

            str:
                The resulting hexadecimal color value
        """
        rgb_256 = tuple((c // 256 for c in self.tk_master.winfo_rgb(color)))
        r, g, b = tuple((c / 256 for c in rgb_256))
        h, s, v = colorsys.rgb_to_hsv(r, g, b)

        # hue
        if h * (1 + hd) > 1:
            h = 1
        elif h * (1 + hd) < 0:
            h = 0
        else:
            h *= 1 + hd
        # saturation
        if s * (1 + sd) > 1:
            s = 1
        elif s * (1 + sd) < 0:
            s = 0
        else:
            s *= 1 + sd
        # value
        if v * (1 + vd) > 1:
            v = 0.95
        elif v * (1 + vd) < 0.05:
            v = 0.05
        else:
            v *= 1 + vd

        rgb_float = colorsys.hsv_to_rgb(h, s, v)
        rgb_int = [int(x * 256) for x in rgb_float]
        return "#{:02x}{:02x}{:02x}".format(*rgb_int)

    def rgb_to_hsv(self, r, g, b):
        """Convert an rgb to hsv color value.

        Parameters:
            r (float):
                red
            g (float):
                green
            b (float):
                blue

        Returns:
            Tuple[float, float, float]: The hsv color value.
        """
        return colorsys.rgb_to_hsv(r, g, b)

    # Color utilities end



# encoding: utf-8
import tkinter as tk
from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKButton
from pygubu.widgets.pathchooserinput import PathChooserInput, PathChooserButton
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class PathChooserBaseMixin:
    base_properties = (
        "type",
        "path",
        "initialdir",
        "mustexist",
        "title",
        "defaultextension",
    )

    def _process_property_value(self, pname, value):
        if pname == "mustexist":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "mustexist":
            return self._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


class PathChooserInputBO(PathChooserBaseMixin, BuilderObject):
    class_ = PathChooserInput
    properties = PathChooserBaseMixin.base_properties + (
        "image",
        "textvariable",
        "state",
    )
    virtual_events = ("<<PathChooserPathChanged>>",)


_builder_id = f"{_plugin_uid}.PathChooserInput"
register_widget(
    _builder_id,
    PathChooserInputBO,
    "PathChooserInput",
    ("ttk", _tab_widgets_label),
)
_old_bid = "pygubu.builder.widgets.pathchooserinput"
register_widget(
    _old_bid,
    PathChooserInputBO,
    public=False,
)


class PathChooserButtonBO(PathChooserBaseMixin, BuilderObject):
    class_ = PathChooserButton
    properties = PathChooserBaseMixin.base_properties + tuple(
        set(TTKButton.properties) - set(("command", "default"))
    )
    virtual_events = ("<<PathChooserPathChanged>>",)


_builder_id = f"{_plugin_uid}.PathChooserButton"
register_widget(
    _builder_id,
    PathChooserButtonBO,
    "PathChooserButton",
    ("ttk", _tab_widgets_label),
)



from typing import Optional
from collections import OrderedDict
from .transformer import NoopTransfomer, TransformationError
from .validation.base import ExecutionContext
from .validation.constraint.form import Form as FormConstraint
from .validation.constraint.form import FormValidator


class FormError(Exception):
    ...


class FormField:
    def __init__(self, widget, **kw):
        self.widget = widget
        self.required = kw.get("required", True)
        self.initial = kw.get("initial", None)
        self.constraints = kw.get("constraints", [])
        self.help = kw.get("help")
        # self.mapped = kw.get("mapped", True)
        noop_transformer = NoopTransfomer()
        self.model_transformer = kw.get("model_transformer", noop_transformer)
        self.view_transformer = kw.get("view_transformer", noop_transformer)

    @property
    def data(self):
        # NOTE: should return initial if field is disabled.
        if self.widget.wis_disabled():
            return self.initial
        return self.view_transformer.reversetransform(self.widget.wget_value())

    @data.setter
    def data(self, value):
        self.widget.wset_value(
            self.view_transformer.transform(
                self.model_transformer.transform(value)
            )
        )

    @property
    def field_name(self):
        return self.widget.field_name

    def clean(self, value):
        """
        Validate the given value and return its "cleaned" value as an
        appropriate Python object. Raise TransormationError for any errors.
        """
        value = self.model_transformer.reversetransform(value)
        return value


class Form:
    def __init__(self):
        self.fields = OrderedDict()
        self.initialized = False
        self.fields_initial = {}
        self.info_display = {}
        self.cleaned_data = {}
        self.transformation_error = None
        self._errors = {}
        self.translator = None

    def add(self, name: str, field: FormField):
        self.fields[name] = field

    def add_info_display(self, name: str, info_display):
        self.info_display[name] = info_display

    def edit(self, data: dict, initial_bag: Optional[dict] = None):
        """Intializes form to edit data values."""
        self.initialized = True
        self.fields_initial = {}
        if initial_bag is None:
            initial_bag = {}
        for name, field in self.fields.items():
            field_initial = initial_bag.get(name, field.initial)
            field_initial = "" if field_initial is None else field_initial
            self.fields_initial[name] = field_initial
            field.data = data.get(name, field_initial)
            self.edit_field_init(field)

    def edit_field_init(self, field):
        if field.field_name in self.info_display:
            field_info = self.info_display[field.field_name]
            field_info.clear()
            if field.help:
                field_info.show_help(field.help)

    def is_valid(self):
        return self.is_bound and not self._errors

    def get_data(self):
        return self.cleaned_data

    def submit(self):
        if self.initialized:
            self.is_bound = True
            self.submit_init()
            self.do_submit()
            self.do_validation()
        else:
            raise FormError(
                "Form initialization error. Call form.edit() before submit"
            )

    def submit_init(self):
        self.cleaned_data = {}
        self._errors = {}
        self.transformation_error = None
        for name, field in self.fields.items():
            field.widget.wmark_invalid(False)
        for name in self.info_display:
            self.info_display[name].clear()

    def do_submit(self):
        for name, field in self.fields.items():
            value = field.initial if field.widget.wis_disabled() else field.data
            try:
                value = field.clean(value)
                self.cleaned_data[name] = value
            except TransformationError as e:
                self.transformation_error = e

    def field_submit_pass(self, field):
        if field.help:
            field_info = self.info_display.get(field.field_name, None)
            if field_info is not None:
                field_info.show_help(field.help)

    def field_submit_error(self, field, error):
        field.widget.wmark_invalid(True)
        field_info = self.info_display.get(field.field_name, None)
        if field_info is not None:
            field_info.show_error(error)

    def do_validation(self):
        constraint = FormConstraint()
        validator = FormValidator()
        context = ExecutionContext(translator=self.translator)
        validator.initialize(context)
        validator.validate(self, constraint)

        for fname, field in self.fields.items():
            if fname in self._errors:
                self.field_submit_error(field, self._errors[fname])
            else:
                self.field_submit_pass(field)

    def add_error(self, field_name, error):
        self._errors[field_name] = error



from pygubu.api.v1 import register_custom_property
from pygubu.plugins.tkmt import _designer_tab_label, _plugin_uid

_builder_uid = f"{_plugin_uid}.*"

tkmt_properties = {
    "args": {
        "editor": "json_entry",
        "help": "Args passsed to command property. In designer, a json list.",
    },
    "col": {"editor": "naturalnumber"},
    "colspan": {"editor": "naturalnumber", "min_value": 1},
    "columnnames": {  # Treeview
        "editor": "json_entry",
        "help": "Column header names. In designer, a json list of strings.",
        "json_type": list,
        "json_item_type": str,
    },
    "columnwidths": {  # Treeview
        "editor": "json_entry",
        "help": ("Width of each column, should be same size as columnnames." 
                 + " In designer, a json list of ints"),
        "json_type": list,
        "json_item_type": int,
    },
    "command": {"editor": "simplecommandentry"},
    "datacolumnnames": {  # Treeview
        "editor": "json_entry",
        "help": """Defaults to columnnames, a mapping of columnnames to the data file.
Should be same size as columnnames. In designer, a json list of strings.""",
        "json_type": list,
        "json_item_type": str,
    },
    "data": {  # Treeview
        "editor": "entry",
        "help": """The tree data.
In designer, use a resource URI here, example:  res://my_treeview_data""",
    },
    "defaulttext": {"editor": "entry"},  # MenuButton
    "disabled": {
        "editor": "choice",
        "values": ("", "True"),
        "state": "readonly",
    },
    "increment": {"editor": "realnumber"},
    "invalidcommand": {
        "editor": "simplecommandentry",
    },
    "invalidcommandargs": {
        "editor": "json_entry",
        "help": "Args passsed to invalidcommand property. In designer, a json list.",
    },
    "lower": {"editor": "realnumber"},
    "makeResizable": {
        "editor": "choice",
        "values": ("", "all", "recursive", "onlyframes"),
        "state": "readonly",
    },
    "menu": {  # MenuButton
        "buid": f"{_plugin_uid}.MenuButton",
        "editor": "entry",
        "help": "Use a resource URI here, example:  res://my_menu",
    },
    "mode": {
        "editor": "choice",
        "values": ("", "light", "dark"),
        "state": "readonly",
    },
    "newframe": {  # Treeview
        "editor": "choice",
        "values": ("", "True", "False"),
        "state": "readonly",
    },
    "openkey": {  # Treeview
        "editor": "alphanumentry",
    },
    "orient": {
        "editor": "choice",
        "values": ("", "horizontal", "vertical"),
        "state": "readonly",
    },
    "padx": {
        "editor": "twodimensionentry",
    },
    "pady": {
        "editor": "twodimensionentry",
    },
    "row": {"editor": "naturalnumber"},
    "rowspan": {"editor": "naturalnumber", "min_value": 1},
    "subentryname": {"editor": "entry"},  # Treeview
    "theme": {
        "editor": "choice",
        "values": ("", "azure", "sun-valley", "park"),
        "state": "readonly",
    },
    "upper": {"editor": "realnumber"},
    "usecommandlineargs": {
        "editor": "choice",
        "values": ("", "True", "False"),
        "state": "readonly",
    },
    "useconfigfile": {
        "editor": "choice",
        "values": ("", "True", "False"),
        "state": "readonly",
    },
    "validate": {
        "editor": "choice",
        "values": (
            "",
            "all",
            "focus",
            "focusin",
            "focusout",
            "key",
            "none",
        ),
        "state": "readonly",
    },
    "validatecommand": {"editor": "simplecommandentry"},
    "validatecommandargs": {
        "editor": "json_entry",
        "help": "Args passsed to validatecommand property. In designer, a json list.",
    },
    "validatecommandmode": {
        "editor": "choice",
        "values": ("", "%P", "%d %i %P %s %S %v %V %W"),
        # "state": "readonly",
    },
    "values": [
        {  # OptionMenu
            "editor": "json_entry",
            "buid": f"{_plugin_uid}.OptionMenu",
            "help": "A json list of strings.",
        },
        {  # Combobox
            "editor": "json_entry",
            "buid": f"{_plugin_uid}.Combobox",
            "help": "A json list of strings.",
        },
        {  # NonnumericalSpinbox
            "editor": "json_entry",
            "buid": f"{_plugin_uid}.NonnumericalSpinbox",
            "help": "A json list of strings.",
        },
    ],
    "variable": [
        {
            "buid": f"{_plugin_uid}.Scale",
            "editor": "tkvarentry",
            "type_choices": ("int", "double"),
            "type_default": "int",
        },
        {
            "buid": f"{_plugin_uid}.NumericalSpinbox",
            "editor": "tkvarentry",
            "type_choices": ("int", "double"),
            "type_default": "int",
        },
        {
            "buid": f"{_plugin_uid}.NonnumericalSpinbox",
            "editor": "tkvarentry",
            "type_choices": ("string",),
            "type_default": "string",
        },
    ],
}

for prop in tkmt_properties:
    definitions = tkmt_properties[prop]
    if isinstance(definitions, dict):
        definitions = [definitions]
    for definition in definitions:
        builder_uid = definition.pop("buid", _builder_uid)
        editor = definition.pop("editor", "entry")
        register_custom_property(builder_uid, prop, editor, **definition)



from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKEntry
from tkcalendar import DateEntry

from ..tkcalendar import _designer_tab_label, _plugin_uid
from .basecalendar import CalendarBaseBO


class DateEntryBO(CalendarBaseBO, TTKEntry):
    class_ = DateEntry
    OPTIONS_CUSTOM = CalendarBaseBO.OPTIONS_CUSTOM + ("calendar_cursor",)
    properties = (
        TTKEntry.OPTIONS_STANDARD + TTKEntry.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = CalendarBaseBO.ro_properties + ("calendar_cursor",)
    virtual_events = ("<<DateEntrySelected>>",)


_builder_id = f"{_plugin_uid}.DateEntry"

register_widget(
    _builder_id,
    DateEntryBO,
    "DateEntry",
    ("ttk", _designer_tab_label),
)



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)

from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid
from pygubu.widgets.fontinput import FontInput


#
# Builder definition section
#


class FontInputBO(BuilderObject):
    class_ = FontInput
    virtual_events = (FontInput.EVENT_FONT_CHANGED,)


_wname = "FontInput"
_builder_id = f"{_plugin_uid}.{_wname}"

register_widget(_builder_id, FontInputBO, _wname, ("ttk", _tab_widgets_label))



# Fix background of file selector dialog on linux

FIX_LINUX_FILE_DIALOG = True


# Constants

TTK_CLAM = "clam"
TTK_ALT = "alt"
TTK_DEFAULT = "default"

PRIMARY = "primary"
SECONDARY = "secondary"
SUCCESS = "success"
INFO = "info"
WARNING = "warning"
DANGER = "danger"
LIGHT = "light"
DARK = "dark"



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_custom_property
from pygubu.plugins.pygubu import _plugin_uid
from pygubu.plugins.pygubu.forms.base import _plugin_forms_uid as forms_uid
from pygubu.widgets.pathchooserinput import PathChooserInput


_builder_all = f"{_plugin_uid}.*"
_AccordionFrame = f"{_plugin_uid}.AccordionFrame"
_AccordionFrameGroup = f"{_plugin_uid}.AccordionFrameGroup"
_CalendarFrame = f"{_plugin_uid}.CalendarFrame"
_CalendarFrame_old = "pygubu.builder.widgets.calendarframe"
_ColorInput = f"{_plugin_uid}.ColorInput"
_Combobox = f"{_plugin_uid}.Combobox"
_Combobox_old = "pygubu.builder.widgets.combobox"
_Dialog = f"{_plugin_uid}.Dialog"
_Dialog_old = "pygubu.builder.widgets.dialog"
_DockFrame = f"{_plugin_uid}.docframe"
_DockPane = f"{_plugin_uid}.docpane"
_DockWidget = f"{_plugin_uid}.dockwidget"
_forms_all = f"{forms_uid}.*"
_forms_Entry = f"{forms_uid}.ttkwidget.Entry"
_forms_PygubuCombobox = f"{forms_uid}.pygubuwidget.PygubuCombobox"
_forms_FrameFormBuilder = f"{forms_uid}.ttkwidget.FrameFormBuilder"
_forms_LabelWidgetInfo = f"{forms_uid}.ttkwidget.LabelWidgetInfo"
_forms_Combobox = f"{forms_uid}.ttkwidget.Combobox"
_HideableFrame = f"{_plugin_uid}.hideableframe"
_PathChoser_all = f"{_plugin_uid}.PathChooser.*"
_PathChoser_old = "pygubu.builder.widgets.pathchooser.*"
_PathChoserButton = f"{_plugin_uid}.PathChooserButton"
_Tooltip = f"{_plugin_uid}.Tooltip"
_Tooltipttk = f"{_plugin_uid}.Tooltipttk"
_FilterableTreeview = f"{_plugin_uid}.FilterableTreeview"
_EditableTreeview = f"{_plugin_uid}.EditableTreeview"

h_values = _(
    "In designer: json list of key, value pairs\n"
    '    Example: [["A", "Option 1 Label"], ["B", "Option 2 Label"]]\n'
    "In code: an iterable of key, value pairs"
)  # combobox
h_keyvariable = _("Tk variable associated to the key value.")  # combobox
h_state = _("Combobox state.")  # combobox
h_modal = _("Determines if dialog is run in normal or modal mode.")  # Dialog
h_weight = _("The weight value for the pane.")  # DockPane
h_mustexist = _(
    "Dialog option. Determines if path must exist for directory and file dialogs."
    + " The default value is True."
)  # PathChooserInput
h_initialdir = _("Dialog option. Sets initial directory.")  # PathChooserInput
h_title = _("Dialog option. Sets dialog title.")  # PathChooserInput
h_defaultextension = _help = _("Dialog option. Sets default file extension.")

plugin_properties = {
    "calendarfg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "calendarbg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "compound": dict(
        buid=_AccordionFrameGroup,
        editor="choice",
        values=(
            "",
            tk.LEFT,
            tk.RIGHT,
            tk.NONE,
        ),
        state="readonly",
    ),
    "defaultextension": dict(
        buid=[_PathChoser_all, _PathChoser_old], help=h_defaultextension
    ),
    "expanded": dict(
        buid=_AccordionFrameGroup,
        editor="choice",
        values=("", "false", "true"),
        state="readonly",
    ),
    "field_name": [
        dict(buid=_forms_all, editor="fieldname_entry"),
        dict(buid=_forms_LabelWidgetInfo, editor="fieldname_selector"),
    ],
    "firstweekday": dict(
        buid=[_CalendarFrame, _CalendarFrame_old],
        editor="choice",
        values=("0", "6"),
        state="readonly",
        default_value="6",
    ),
    "grouped": dict(
        buid=_DockWidget,
        editor="choice",
        values=("", "true", "false"),
        state="readonly",
    ),
    "headerbg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "headerfg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "height": [
        dict(buid=_AccordionFrame, editor="dimensionentry", default_value=200),
        dict(buid=_HideableFrame, editor="dimensionentry", default_value=200),
        dict(
            buid=_forms_FrameFormBuilder,
            editor="dimensionentry",
            default_value=200,
        ),
        dict(
            buid=[_Dialog, _Dialog_old],
            editor="dimensionentry",
            default_value=200,
        ),
    ],
    "image": dict(
        buid=[_PathChoser_all, _PathChoser_old],
        editor="imageentry",
        help=_("Image for the button."),
    ),
    "img_expand": dict(editor="imageentry"),
    "img_collapse": dict(editor="imageentry"),
    "initialdir": dict(
        buid=[_PathChoser_all, _PathChoser_old], help=h_initialdir
    ),
    "keyvariable": dict(
        buid=[_Combobox, _Combobox_old, _forms_PygubuCombobox],
        editor="tkvarentry",
        help=h_keyvariable,
    ),
    "label": dict(buid=_AccordionFrameGroup),
    "linewidth": dict(
        buid=[_CalendarFrame, _CalendarFrame_old],
        editor="choice",
        values=("1", "2", "3", "4"),
        state="readonly",
        default_value="1",
    ),
    "mask": dict(buid=f"{_plugin_uid}.Floodgauge"),
    "markbg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "markfg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "modal": [
        dict(
            buid=[_Dialog, _Dialog_old],
            editor="choice",
            values=("true", "false"),
            state="readonly",
            help=h_modal,
        ),
    ],
    "month": dict(
        buid=[_CalendarFrame, _CalendarFrame_old],
        editor="choice",
        values=(
            "",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
        ),
        state="readonly",
    ),
    "mustexist": dict(
        buid=[_PathChoser_all, _PathChoser_old],
        editor="choice",
        values=("", "true", "false"),
        state="readonly",
        default_value="true",
        help=h_mustexist,
    ),
    "path": dict(
        buid=[_PathChoser_all, _PathChoser_old],
        help=_("Initial path value."),
    ),
    "selectbg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "selectfg": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="colorentry"
    ),
    "state": [
        dict(
            buid=[_CalendarFrame, _CalendarFrame_old],
            editor="choice",
            values=("", "normal", "disabled"),
            state="readonly",
        ),
        dict(
            buid=[_Combobox, _Combobox_old],
            editor="choice",
            values=("", "normal", "disabled"),
            state="readonly",
            help=h_state,
        ),
        dict(
            buid=_forms_PygubuCombobox,
            editor="choice",
            values=("", "normal", "disabled"),
            state="readonly",
            help=h_state,
        ),
        dict(
            buid=[_PathChoser_all, _PathChoser_old],
            editor="choice",
            values=("", "normal", "disabled", "readonly"),
            state="readonly",
            help=_("Path entry state."),
        ),
        dict(
            buid=[_forms_Entry, _forms_Combobox],
            editor="choice",
            values=("", "normal", "disabled", "readonly"),
            state="readonly",
        ),
    ],
    "style": dict(
        buid=_AccordionFrameGroup,
        editor="ttkstylechoice",
        default_value="Toolbutton",
    ),
    "text": dict(
        buid=[_Tooltip, _Tooltipttk],
        editor="text",
        help=_("Text to display in tooltip."),
    ),
    "textvariable": [
        dict(
            buid=[_PathChoserButton],
            editor="tkvarentry",
            help=_(
                "Tk variable whose value will be used in place of the text resource."
            ),
        ),
        dict(
            buid=[_PathChoser_all, _PathChoser_old],
            editor="tkvarentry",
            help=_("Tk variable associated to the path property."),
        ),
    ],
    "title": dict(buid=[_PathChoser_all, _PathChoser_old], help=h_title),
    "type": dict(
        buid=[_PathChoser_all, _PathChoser_old],
        editor="choice",
        values=(PathChooserInput.FILE, PathChooserInput.DIR),
        state="readonly",
        default_value=PathChooserInput.FILE,
        help=_("Dialog type"),  # PathChooserInput
    ),
    "value": dict(buid=_ColorInput, editor="colorentry"),
    "values": [
        dict(
            buid=[_Combobox, _Combobox_old, _forms_PygubuCombobox],
            help=h_values,
        ),
    ],
    "weight": [
        dict(buid=_DockPane, editor="integernumber", help=h_weight),
        dict(
            buid=_DockWidget,
            editor="integernumber",
            help=_("The weight value of the widget in the pane"),
        ),
    ],
    "width": [
        dict(buid=_AccordionFrame, editor="dimensionentry", default_value=200),
        dict(buid=_HideableFrame, editor="dimensionentry", default_value=200),
        dict(
            buid=_forms_FrameFormBuilder,
            editor="dimensionentry",
            default_value=200,
        ),
        dict(buid=_PathChoserButton, editor="integernumber"),
        dict(
            buid=[_Dialog, _Dialog_old],
            editor="dimensionentry",
            default_value=200,
        ),
    ],
    "xscrollcommand": dict(
        buid=[_FilterableTreeview, _EditableTreeview],
        editor="scrollsetcommandentry",
    ),
    "year": dict(
        buid=[_CalendarFrame, _CalendarFrame_old], editor="naturalnumber"
    ),
    "yscrollcommand": dict(
        buid=[_FilterableTreeview, _EditableTreeview],
        editor="scrollsetcommandentry",
    ),
}

for prop in plugin_properties:
    definitions = plugin_properties[prop]
    if isinstance(definitions, dict):
        definitions = [definitions]
    for definition in definitions:
        builders = definition.pop("buid", _builder_all)
        if isinstance(builders, str):
            builders = [builders]
        editor = definition.pop("editor", "entry")
        for builder_uid in builders:
            register_custom_property(builder_uid, prop, editor, **definition)



import pygubu.plugins.pygubu.designer.properties
from pygubu.api.v1 import IDesignerPlugin
from pygubu.utils.widget import crop_widget
from pygubu.stockimage import StockRegistry, StockImageCache, StockImage
from .toplevelframe import ToplevelFramePreviewBO
from .basehelpers import TKTextPreviewBO


class PygubuDesignerPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid in (
            "tk.Tk",
            "tk.Toplevel",
            "pygubu.widgets.Dialog",
            "pygubu.builder.widgets.dialog",  # This will be removed
        ):
            return ToplevelFramePreviewBO
        elif builder_uid == "tk.Text":
            return TKTextPreviewBO
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        top = None
        if builder_uid == "tk.Toplevel":
            top = builder.get_object(widget_id, top_master)
        elif builder_uid == "tk.Tk":
            # for a new tk root, create a diferent image cache:
            def on_root_created(root):
                image_cache = StockImageCache(root, StockImage.registry)
                builder.image_cache = image_cache

            builder.on_first_object = on_root_created
            top = builder.get_object(widget_id)
        elif builder_uid == "pygubu.builder.widgets.dialog":
            dialog = builder.get_object(widget_id, top_master)
            dialog.run()
            top = dialog.toplevel
        return top

    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""
        if builder_uid.endswith(".pathchooserinput"):
            crop_widget(widget.entry)
            crop_widget(widget.folder_button)
        if builder_uid.endswith(".ColorInput"):
            crop_widget(widget._entry)
            crop_widget(widget._button)

    def ensure_visibility_in_preview(self, builder, selected_uid: str):
        """Ensure visibility of selected_uid in preview.
        Usage example:
            Activate a tab of a Notebook if the selected widget is
            inside the notebook.
        """
        xpath = ".//object[@class='ttk.Notebook.Tab']"
        # find all tabs
        tabs = builder.uidefinition.root.findall(xpath)
        if tabs is None:
            return

        for tab in tabs:
            tab_id = tab.get("id")
            # Check if this tab was clicked
            if tab_id == selected_uid:
                xpath = "./child/object[1]"
                child = tab.find(xpath)
                # A tab can be empty, check that.
                if child is not None:
                    child_id = child.get("id")
                    notebook = builder.objects[tab_id].widget
                    current_tab = builder.objects[child_id].widget
                    notebook.select(current_tab)
                    notebook.update()
                    # Found, stop searching
                    break
            # check if selected_uid is inside this tab
            xpath = f".//object[@id='{selected_uid}']"
            o = tab.find(xpath)
            if o is not None:
                # selected_uid is inside, find the tab child
                # and select this tab
                xpath = "./child/object[1]"
                child = tab.find(xpath)
                child_id = child.get("id")
                notebook = builder.objects[tab_id].widget
                current_tab = builder.objects[child_id].widget
                notebook.select(current_tab)
                notebook.update()
                # Found, stop searching
                break



from tkcalendar import Calendar, DateEntry
from pygubu.plugins.tkcalendar.calendar import CalendarBO
from pygubu.plugins.tkcalendar.dateentry import DateEntryBO


class CalendarPreview(Calendar):
    """
    Temporary fix for recreating theme on every <<ThemeChanged>> event.
    """

    ...

    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)

        self.theme_change_cbid = None
        self._theme_name = None
        self._setup_style_fixed()

        self.bind("<<ThemeChanged>>", self.schedule_style_update)

    def _setup_style(self, event=None):
        # Setup is ignored here on purpose.
        return None

    def _setup_style_fixed(self, event=None):
        super()._setup_style(event)
        self._theme_name = self.style.theme_use()

    def schedule_style_update(self, event=None):
        if self.theme_change_cbid is None:
            self.theme_change_cbid = self.after(10, self._on_theme_change)

    def _on_theme_change(self):
        theme = self.style.theme_use()
        if self._theme_name != theme:
            # the theme has changed, update the DateEntry style to look like a combobox
            self._theme_name = theme
            self._setup_style_fixed()
        self.theme_change_cbid = None


class DateEntryPreview(DateEntry):
    """
    Temporary fix for recreating theme on every <<ThemeChanged>> event.
    """

    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self.theme_change_cbid = None
        self.bind("<<ThemeChanged>>", self.schedule_style_update)

    def _setup_style(self, event=None):
        super()._setup_style(event)
        self._theme_name = self.style.theme_use()

    def schedule_style_update(self, event=None):
        if self.theme_change_cbid is None:
            self.theme_change_cbid = self.after(10, self._on_theme_change)

    def _on_theme_change(self):
        theme = self.style.theme_use()
        if self._theme_name != theme:
            # the theme has changed, update the DateEntry style to look like a combobox
            self._theme_name = theme
            self._setup_style()
        self.theme_change_cbid = None


class CalendarPreviewBO(CalendarBO):
    class_ = CalendarPreview


class DateEntryPreviewBO(DateEntryBO):
    class_ = DateEntryPreview



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support
from unittest.case import SkipTest

has_tkcalendar = True
try:
    from tkcalendar import Calendar, DateEntry
except ImportError:
    has_tkcalendar = False


class TestTkcalendarCalendar(unittest.TestCase):
    def setUp(self):
        if not has_tkcalendar:
            raise SkipTest("tkcalendar not installed")
        
        support.root_deiconify()
        xmldata = "test_plugin_tkcalendar.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("frame1")
        self.calendar = builder.get_object("calendar")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.calendar, Calendar)
        self.widget.destroy()


class TestTkcalendarDateEntry(unittest.TestCase):
    def setUp(self):
        if not has_tkcalendar:
            raise SkipTest("tkcalendar not installed")
        
        support.root_deiconify()
        xmldata = "test_plugin_tkcalendar.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("frame2")
        self.entry = builder.get_object("dateentry")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.entry, DateEntry)
        self.widget.destroy()



import importlib
from typing import Optional
from pygubu.i18n import _
from pygubu.api.v1 import BuilderLoaderPlugin, IDesignerPlugin


_designer_tab_label = _("tkintertable")
_plugin_uid = "tkintertable"


class StandardTKWidgetsLoader(BuilderLoaderPlugin, IDesignerPlugin):
    _module = "pygubu.plugins.tkintertable.table"

    #
    # IPluginBase interface methods
    #
    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("tkintertable")
        return spec is not None

    def get_designer_plugin(self) -> Optional[IDesignerPlugin]:
        return self

    #
    # IBuilderLoaderPlugin interface methods
    #
    def get_module_for(self, identifier: str) -> str:
        return self._module

    def get_all_modules(self):
        return (self._module,)

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("tkintertable.")

    #
    # IDesignerPlugin interface methods
    #
    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""
        if builder_uid.endswith(f"{_plugin_uid}.TableCanvas"):

            def _no_op(event=None):
                pass

            seqlist = (
                "<Button-1>",
                "<Button-2>",
                "<Button-3>",
                "<Button-4>",
                "<Button-5>",
                "<MouseWheel>" "<Double-Button-1>",
                "<Control-Button-1>",
                "<Shift-Button-1>",
                "<B1-Motion>",
            )
            for seq in seqlist:
                widget.bind(seq, _no_op)



import sys
import os
import json
import tkinter as tk
import TKinterModernThemes as tkmt
from pygubu.plugins.tkmt.widgets import ThemedTKinterFrameBO
from pygubu.api.v1 import register_builder


class ThemedTKinterFrameTLPreview(tkmt.ThemedTKinterFrame):
    def __init__(
        self,
        title: str,
        theme: str = "",
        mode: str = "",
        usecommandlineargs=True,
        useconfigfile=True,
    ):

        self.root: tk.Tk = tk.Tk()
        self.root.protocol("WM_DELETE_WINDOW", self.handleExit)
        self.root.title(title)

        # region Set Theme
        if usecommandlineargs:
            args = sys.argv
            if len(args) == 3:
                theme = args[1]
                mode = args[2]

        if useconfigfile:
            try:
                with open("themeconfig.json") as f:
                    themeconfig = json.load(f)
                    if "theme" in themeconfig:
                        theme = themeconfig["theme"]
                    if "mode" in themeconfig:
                        mode = themeconfig["mode"]
            except (FileNotFoundError, json.JSONDecodeError):
                pass  # no config file was specified

        if theme == "":
            theme = "park"

        if mode == "":
            mode = "dark"

        try:
            path = os.path.abspath(
                tkmt.__file__
                + "/../themes/"
                + theme.lower()
                + "/"
                + theme.lower()
                + ".tcl"
            )
            self.root.tk.call("source", path)
        except tk.TclError:
            pass  # theme already loaded...
        self.root.tk.call("set_theme", mode.lower())

        self.theme = theme.lower()
        self.mode = mode.lower()
        # endregion

        super(tkmt.ThemedTKinterFrame, self).__init__(self.root, "Master Frame")

    def handleExit(self):
        self.root.withdraw()

    def destroy(self):
        self.root.destroy()


class ThemedTKinterFrameTLPreviewBO(ThemedTKinterFrameBO):
    class_ = ThemedTKinterFrameTLPreview


register_builder(
    "tkmt.ThemedTKinterFrameTLPreview",
    ThemedTKinterFrameTLPreviewBO,
    "ThemedTKinterFrameTLPreview",
    ("ttk",),
    public=False,
)


if __name__ == "__main__":
    test = ThemedTKinterFrameTLPreview("Preview Test")
    test.run()



from pygubu.api.v1 import IDesignerPlugin
from .properties import _calendar, _dateentry
from .previewhelper import CalendarPreviewBO, DateEntryPreviewBO


class TkcalendarPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid == _calendar:
            return CalendarPreviewBO
        elif builder_uid == _dateentry:
            return DateEntryPreviewBO
        return None



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestCustomWidget(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_custom_widget.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")
        self.custom_widget = builder.get_object("custom_widget")

    def tearDown(self):
        support.root_withdraw()

    def test_loading(self):
        message = self.custom_widget.get_message()
        self.assertEqual(message, "CustomLabel")
        self.widget.destroy()



from pygubu.i18n import _
from pygubu.api.v1 import register_custom_property
from pygubu.plugins.awesometkinter import _plugin_uid


_builder_all = f"{_plugin_uid}.*"
_button3d = f"{_plugin_uid}.Button3d"
_radiobutton = f"{_plugin_uid}.Radiobutton"
_checkbutton = f"{_plugin_uid}.Checkbutton"
_frame3d = f"{_plugin_uid}.Frame3d"
_scrollableframe = f"{_plugin_uid}.ScrollableFrame"
_autowraplabel = f"{_plugin_uid}.AutoWrappingLabel"
_autofitlabel = f"{_plugin_uid}.AutofitLabel"
_radialpbar = f"{_plugin_uid}.RadialProgressbar"
_radialpbar3d = f"{_plugin_uid}.RadialProgressbar3d"
_segmentbar = f"{_plugin_uid}.Segmentbar"
_simplesbar = f"{_plugin_uid}.SimpleScrollbar"
_scrolledtext = f"{_plugin_uid}.ScrolledText"


plugin_properties = {
    "autoscroll": [
        dict(
            buid=_scrollableframe,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("auto scroll to bottom if new items added to frame"),
        ),
        dict(
            buid=_scrolledtext,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("automatic vertical scrolling"),
        ),
    ],
    "base_img": dict(
        buid=_radialpbar,
        editor="imageentry",
        help=_("base image for progressbar"),
    ),
    "check_mark_color": dict(
        buid=_checkbutton,
        editor="colorentry",
    ),
    "bd": dict(
        buid=_scrolledtext, editor="naturalnumber", help=_("border width")
    ),
    "bg": [
        dict(buid=_button3d, editor="colorentry", help=_("button color")),
        dict(
            buid=_radiobutton,
            editor="colorentry",
            help=_('background color "should match parent bg"'),
        ),
        dict(buid=_frame3d, editor="colorentry", help=_("color of frame")),
        dict(
            buid=_scrollableframe,
            editor="colorentry",
            help=_("background color"),
        ),
        dict(
            buid=_radialpbar, editor="colorentry", help=_("color of base ring")
        ),
        dict(buid=[_segmentbar, _simplesbar], editor="colorentry"),
        dict(
            buid=_scrolledtext, editor="colorentry", help=_("background color")
        ),
    ],
    "box_color": dict(
        buid=_checkbutton,
        editor="colorentry",
    ),
    "fg": [
        dict(
            buid=[_button3d, _radiobutton],
            editor="colorentry",
            help=_("text color"),
        ),
        dict(
            buid=_radialpbar,
            editor="colorentry",
            help=_("color of indicator ring"),
        ),
        dict(buid=_radialpbar3d, editor="colorentry"),
        dict(buid=_segmentbar, editor="colorentry"),
        dict(
            buid=_scrolledtext, editor="colorentry", help=_("foreground color")
        ),
    ],
    "font": [
        dict(buid=_radiobutton, editor="fontentry"),
        dict(
            buid=_radialpbar,
            editor="fontentry",
            help=_("tkinter font for percentage text"),
        ),
    ],
    "font_size_ratio": dict(
        buid=_radialpbar,
        editor="spinbox",
        from_=0.1,
        to=1,
        increment=0.1,
        help=_(
            "font size to progressbar width ratio,"
             + "e.g. for a progressbar size 100 pixels, a 0.1 ratio means font size 10"
        ),
    ),
    "height": [
        dict(buid=_segmentbar, editor="dimensionentry"),
        dict(
            buid=[_frame3d, _scrollableframe],
            editor="dimensionentry",
            default_value=200,
        ),
    ],
    "hbar_width": [
        dict(
            buid=_scrollableframe,
            editor="dimensionentry",
            help=_("horizontal scrollbar width"),
        ),
        dict(
            buid=_scrolledtext,
            editor="naturalnumber",
            help=_("horizontal scrollbar width"),
        ),
    ],
    "hscroll": [
        dict(
            buid=_scrollableframe,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("use horizontal scrollbar"),
        ),
        dict(
            buid=_scrolledtext,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("include horizontal scrollbar"),
        ),
    ],
    "ind_bg": dict(
        buid=_radiobutton,
        editor="colorentry",
        help=_('indicator ring background "fill color"'),
    ),
    "ind_mark_color": dict(
        buid=_radiobutton, editor="colorentry", help=_("check mark color")
    ),
    "ind_outline_color": dict(
        buid=_radiobutton,
        editor="colorentry",
        help=_("indicator outline / ring color"),
    ),
    "indicator_img": dict(
        buid=_radialpbar,
        editor="imageentry",
        help=_("indicator image for progressbar"),
    ),
    "max_chars": dict(
        buid=_scrolledtext,
        editor="naturalnumber",
        help=_(
            "maximum characters allowed in Text widget, "
             + "text will be truncated from the beginning to match the max chars"
        ),
    ),
    "parent_bg": dict(
        buid=_radialpbar,
        editor="colorentry",
        help=_("color of parent container"),
    ),
    "refresh_time": dict(
        buid=_autofitlabel,
        editor="naturalnumber",
        help=_("milliseconds"),
    ),
    "sbar_bg": dict(
        buid=[_scrollableframe, _scrolledtext],
        editor="colorentry",
        help=_("color of scrollbars' trough, default to frame's background"),
    ),
    "sbar_fg": dict(
        buid=[_scrollableframe, _scrolledtext],
        editor="colorentry",
        help=_("color of scrollbars' slider"),
    ),
    "slider_color": dict(buid=_simplesbar, editor="colorentry"),
    "text_color": dict(
        buid=_checkbutton,
        editor="colorentry",
    ),
    "text_bg": dict(buid=_radialpbar3d, editor="colorentry"),
    "text_fg": [
        dict(
            buid=_radialpbar,
            editor="colorentry",
            help=_("percentage text color"),
        ),
        dict(buid=_radialpbar3d, editor="colorentry"),
    ],
    "vscroll": [
        dict(
            buid=_scrollableframe,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("use vertical scrollbar"),
        ),
        dict(
            buid=_scrolledtext,
            editor="choice",
            values=("", "false", "true"),
            state="readonly",
            help=_("include vertical scrollbar"),
        ),
    ],
    "vbar_width": [
        dict(
            buid=_scrollableframe,
            editor="dimensionentry",
            help=_("vertical scrollbar width"),
        ),
        dict(
            buid=_scrolledtext,
            editor="naturalnumber",
            help=_("vertical scrollbar width"),
        ),
    ],
    "width": [
        dict(buid=[_segmentbar, _simplesbar], editor="naturalnumber"),
        dict(
            buid=[_frame3d, _scrollableframe],
            editor="dimensionentry",
            default_value=200,
        ),
    ],
}

for prop in plugin_properties:
    definitions = plugin_properties[prop]
    if isinstance(definitions, dict):
        definitions = [definitions]
    for definition in definitions:
        builders = definition.pop("buid", _builder_all)
        if isinstance(builders, str):
            builders = [builders]
        editor = definition.pop("editor", "entry")
        for builder_uid in builders:
            register_custom_property(builder_uid, prop, editor, **definition)



#!/usr/bin/python3
import tkinter as tk
import tkinter.colorchooser
import tkinter.ttk as ttk
from pygubu.utils.widget import WidgetConfigureMixin
from pygubu.widgets.colorinputui import ColorInputUI


#
# Manual user code
#


class ColorInput(WidgetConfigureMixin, ColorInputUI):
    BGCOLOR = ""
    EVENT_COLOR_CHANGED = "<<ColorInput:ColorChanged>>"
    KEY_PRESS_MS = 850

    def __init__(self, master=None, **kw):
        entry_var = kw.pop("textvariable", None)
        super().__init__(master, **kw)
        entry_var = tk.StringVar(value="") if entry_var is None else entry_var
        self._entry_var = entry_var
        self._entry.configure(textvariable=self._entry_var)
        self._color = ""
        self._kp_cb = None

        style = ttk.Style(master)
        cls = type(self)
        cls.BGCOLOR = style.lookup("TFrame", "background")

    def _widget_cget(self, option):
        if option == "value":
            return self._color
        if option == "textvariable":
            return self._entry_var
        return super()._widget_cget(option)

    def _configure_get(self, option):
        if option in ("value", "textvariable"):
            return self._widget_cget(option)
        return super()._configure_get(option)

    def _configure_set(self, **kw):
        key = "textvariable"
        if key in kw:
            value = kw.pop(key)
            self._entry_var = value
            self._entry.configure(textvariable=value)
        key = "value"
        if key in kw:
            value = kw.pop(key)
            self._set_value(value)
        key = "state"
        if key in kw:
            state = kw.pop(key)
            if state == "readonly":
                self._entry.configure(state=state)
                self._button.configure(state="disabled")
            else:
                self._entry.configure(state=state)
                self._button.configure(state=state)
        return super()._configure_set(**kw)

    def _set_value(self, txtcolor):
        self._color = txtcolor
        self._entry_var.set(txtcolor)
        self._show_color(txtcolor)

    def on_focusout(self, event=None):
        self.validate_change(self._entry_var.get())

    def on_keypress_after(self):
        self.validate_change(self._entry_var.get())

    def on_keypress(self, event=None):
        if self._kp_cb is not None:
            self.after_cancel(self._kp_cb)
        self._kp_cb = self.after(self.KEY_PRESS_MS, self.on_keypress_after)

    def is_color(self, txtcolor):
        try:
            self.winfo_rgb(txtcolor)
            return True
        except tk.TclError:
            pass
        return False

    def on_picker_clicked(self):
        current = self._entry_var.get()
        current = current if self.is_color(current) else None
        txtcolor = None
        _, txtcolor = tk.colorchooser.askcolor(current, parent=self)
        if txtcolor is not None:
            self._set_value(txtcolor)
            self.event_generate(self.EVENT_COLOR_CHANGED)

    def validate_change(self, newcolor):
        if self.is_color(newcolor) or newcolor == "":
            self._set_value(newcolor)
            self.event_generate(self.EVENT_COLOR_CHANGED)

    def _show_color(self, newcolor):
        newcolor = newcolor if newcolor else self.BGCOLOR
        self._frame.configure(background=newcolor)


if __name__ == "__main__":
    root = tk.Tk()
    widget = ColorInput(root)
    widget.pack(expand=True, fill="both")

    def on_color_changed(event):
        value = event.widget.cget("value")
        print(f"Color changed to: {value}")

    widget.bind(ColorInput.EVENT_COLOR_CHANGED, on_color_changed)
    widget.configure(value="brown")

    root.mainloop()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


class Constraint(ABC):
    CLASS_CONSTRAINT = 1
    PROPERTY_CONSTRAINT = 2

    def __init__(self, *args, options=None, groups=None, payload=None):
        self.groups = groups
        self.payload = payload
        self.options = options

    @abstractmethod
    def validated_by(self):
        ...

    def get_targets(self):
        return self.PROPERTY_CONSTRAINT


class ConstraintViolation:
    def __init__(self, message, constraint, *, code=None, params=None):
        self._message = message
        self.code = code
        self.params = params
        self.constraint = constraint

    def _msg_format(self):
        if self.params is None:
            return self._message
        else:
            return self._message % self.params

    @property
    def message(self):
        return self._msg_format()

    def messages(self):
        yield self.code, self.message


class ConstraintViolationList(ConstraintViolation):
    def __init__(self, violations: list):
        self.violations = violations

    @property
    def message(self):
        msgs = [violation.message for violation in self.violations]
        return "\n".join(msgs)

    def messages(self):
        for violation in self.violations:
            yield violation.code, violation.message

    def __bool__(self):
        return bool(self.violations)


class ExecutionContext:
    def __init__(self, translator=None):
        self._violations = []
        self._translator = (
            self._noop_trans if translator is None else translator
        )

    def _noop_trans(self, s):
        """Default no-op translator"""
        return s

    @property
    def violations(self):
        return ConstraintViolationList(self._violations)

    def add_violation(self, /, **kw):
        msg_key = "message"
        if msg_key in kw:
            kw[msg_key] = str(self._translator(kw[msg_key]))
        violation = ConstraintViolation(**kw)
        self._violations.append(violation)

    def clear(self):
        self._violations = []


class ConstraintValidator(ABC):
    def __init__(self):
        self.context: ExecutionContext = None

    def initialize(self, context: ExecutionContext):
        self.context = context

    @abstractmethod
    def validate(self, value, constraint):
        ...



# encoding: UTF-8
import tkinter as tk
import tkinter.ttk as ttk


img_expand = """\
R0lGODlhEAAQAIAAAAAAAAAAACH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAAEALAAAAAAQABAA
AAIXjI+py+0P4wK0WprunRo0/VgRJpXmyRQAOw==
"""

img_collapse = """\
R0lGODlhEAAQAIAAAAAAAAAAACH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAAEALAAAAAAQABAA
AAIdjI+pywGtwINHTmpvy3rxnnwQh1mUI52o6rCuWgAAOw==
"""


class AccordionFrame(ttk.Frame):
    """An accordion like widget.
    Usage:
        acframe = AccordionFrame(master)
        acframe.grid()
        g = acframe.add_group('g1', 'Group1')
        label = ttk.Label(g, text='Label on group1')
        label.grid()
    """

    EXPAND = 1
    COLLAPSE = 2

    def __init__(self, master=None, **kw):
        ttk.Frame.__init__(self, master, **kw)
        self.__images = None
        self.__groups = {}
        self.columnconfigure(0, weight=1)
        self.img_expand = None
        self.img_collapse = None

    def _get_image(self, imgid):
        if imgid == self.EXPAND:
            if self.img_expand is None:
                self.img_expand = tk.PhotoImage(data=img_expand, master=self)
            return self.img_expand
        else:
            if self.img_collapse is None:
                self.img_collapse = tk.PhotoImage(
                    data=img_collapse, master=self
                )
            return self.img_collapse

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "img_expand"
        if key in kw:
            value = kw.pop(key)
            self.img_expand = value
        key = "img_collapse"
        if key in kw:
            value = kw.pop(key)
            self.img_collapse = value
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        option = "img_expand"
        if key == option:
            return self.img_expand
        option = "img_collapse"
        if key == option:
            return self.img_collapse
        return super().cget(key)

    def add_group(
        self,
        gid,
        label=None,
        expanded=True,
        style="Toolbutton",
        compound="left",
    ):

        glabel = label
        if label is None:
            glabel = str(gid)

        # button creation
        btn = ttk.Button(
            self,
            text=glabel,
            style=style,
            image=self._get_image(self.EXPAND),
            compound=compound,
        )
        btn.grid(sticky=tk.EW)
        btn.configure(command=lambda: self._toggle_group(gid))

        # frame creation
        frame = ttk.Frame(self, width=100, height=100)
        frame.grid(sticky=tk.NSEW)
        frame.rowconfigure(0, weight=1)
        frame.columnconfigure(0, weight=1)

        if not expanded:
            self.after_idle(lambda: self._toggle_group(gid))

        # store button, and frame
        self.__groups[gid] = (btn, frame)

        return frame

    def get_group(self, gid):
        return self.__groups[gid][1]

    def group_toogle(self, gid):
        self.__button_clicked(gid)

    def group_config(self, gid, **kw):
        btn, frame = self.__groups[gid]
        label = "label"
        if label in kw:
            btn.configure(text=kw[label])
        expanded = "expanded"
        if expanded in kw:
            self.after_idle(
                lambda: self._toggle_group(gid, toggle_to=kw[expanded])
            )
        compound = "compound"
        if compound in kw:
            btn.configure(compound=kw[compound])
        style = "style"
        if style in kw:
            btn.configure(style=kw[style])

    def _toggle_group(self, gid, toggle_to=None):
        btn, frame = self.__groups[gid]
        if frame.winfo_viewable() or toggle_to is False:
            btn.configure(image=self._get_image(self.COLLAPSE))
            frame.grid_remove()
        else:
            btn.configure(image=self._get_image(self.EXPAND))
            frame.grid()
        self.event_generate("<<AccordionGroupToggle>>")


if __name__ == "__main__":
    root = tk.Tk()

    app = AccordionFrame(root)
    app.grid(sticky=tk.NSEW)

    top = app.winfo_toplevel()
    top.rowconfigure(0, weight=1)
    top.columnconfigure(0, weight=1)

    g = app.add_group("g1", "Tk widgets")
    w = tk.Label(g, text="Label1")
    w.grid()
    w = tk.Label(g, text="Label2")
    w.grid()
    g = app.add_group("g2", "Ttk widgets")
    ##
    app = AccordionFrame(g)
    app.grid(sticky="nsew", padx="5 0")
    g = app.add_group("g1", "Containers")
    w = tk.Label(g, text="Label1")
    w.grid()
    w = tk.Label(g, text="Label2")
    w.grid()
    g = app.add_group("g2", "Control")
    w = tk.Label(g, text="Label3")
    w.grid()
    w = tk.Label(g, text="Label4")
    w.grid()

    tk.mainloop()



import tkinter as tk
import pygubu.plugins.pygubu.combobox_bo as cbb
import pygubu.plugins.pygubu.fontinputbo as fib
import pygubu.forms.pygubuwidget as pygubuwidget

from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
    copy_custom_property,
)
from .base import (
    WidgetBOMixin,
    _plugin_forms_uid,
    _tab_form_widgets_label,
)


_plugin_uid = f"{_plugin_forms_uid}.pygubuwidget"
_designer_tabs = ("tk", "ttk", _tab_form_widgets_label)


class PygubuComboboxBO(WidgetBOMixin, cbb.ComboboxBO):
    class_ = pygubuwidget.PygubuCombobox
    properties = cbb.ComboboxBuilder.properties + WidgetBOMixin.base_properties
    ro_properties = (
        cbb.ComboboxBuilder.ro_properties + WidgetBOMixin.base_properties
    )


_builder_uid = f"{_plugin_uid}.PygubuCombobox"
register_widget(
    _builder_uid,
    PygubuComboboxBO,
    "PygubuCombobox",
    _designer_tabs,
)


class FontInputFWBO(WidgetBOMixin, fib.FontInputBO):
    class_ = pygubuwidget.FontInputFW
    properties = fib.FontInputBO.properties + WidgetBOMixin.base_properties
    ro_properties = (
        fib.FontInputBO.ro_properties + WidgetBOMixin.base_properties
    )


_wname = "FontInput"
_builder_uid = f"{_plugin_uid}.{_wname}"
register_widget(
    _builder_uid,
    FontInputFWBO,
    _wname,
    _designer_tabs,
)



import tkinter as tk
import tkinter.ttk as ttk
import pygubu.plugins.tkmt.widgets as tkmt_builders

from TKinterModernThemes.WidgetFrame import WidgetFrame
from ..base import tkmt_to_tkwidget


class ThemedTkFramePreview(ttk.Frame):
    def __init__(
        self,
        master,
        title: str,
        theme: str = "",
        mode: str = "",
        usecommandlineargs=True,
        useconfigfile=True,
    ):
        super().__init__(master, width=200, height=200)
        self.tkmt_widget = WidgetFrame(self, "Master Frame")


class PreviewBaseMixin:
    def realize(self, parent, extra_init_args: dict = None):
        if self.class_ == ThemedTkFramePreview:
            self.widget = super().realize(parent, extra_init_args)
            self.tkmt_widget = self.widget.tkmt_widget
        else:
            self.tkmt_widget = super().realize(parent, extra_init_args)
            self.widget = tkmt_to_tkwidget(self.tkmt_widget)
        return self.widget

    def get_child_master(self):
        return self.tkmt_widget


class ThemedTKinterFramePreviewBO(
    PreviewBaseMixin, tkmt_builders.ThemedTKinterFrameBO
):
    class_ = ThemedTkFramePreview
    layout_required = True
    pos_args = (
        "master",
        "title",
    )
    properties = pos_args + tkmt_builders.ThemedTKinterFrameBO.kw_args

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"master": master, "title": self.wmeta.identifier}


class FramePreviewBO(PreviewBaseMixin, tkmt_builders.FrameBO):
    ...


class LabelFramePreviewBO(PreviewBaseMixin, tkmt_builders.LabelFrameBO):
    ...


class NotebookPreviewBO(PreviewBaseMixin, tkmt_builders.NotebookBO):
    ...


class NotebookTabPreviewBO(PreviewBaseMixin, tkmt_builders.NotebookTabBO):
    def _set_property(self, target_widget, pname, value):
        return super()._set_property(self.tkmt_widget, pname, value)

    def configure_children(self, target=None):
        super().configure_children(self.tkmt_widget)


class PanedWindowPreviewBO(PreviewBaseMixin, tkmt_builders.PanedWindowBO):
    ...


class PanedWindowPanePreviewBO(
    PreviewBaseMixin, tkmt_builders.PanedWindowPaneBO
):
    ...


class FrameNextColPreviewBO(PreviewBaseMixin, tkmt_builders.FrameNextColBO):
    ...



# encoding: utf-8
import tkinter as tk

from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.widgets.scrolledframe import ScrolledFrame
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class TTKScrolledFrameBO(BuilderObject):
    class_ = ScrolledFrame
    container = True
    container_layout = True
    #    maxchildren = 1
    #    allowed_children = ('tk.Frame', 'ttk.Frame' )
    OPTIONS_STANDARD = ("class_", "cursor", "takefocus", "style")
    OPTIONS_SPECIFIC = ("borderwidth", "relief", "padding", "height", "width")
    OPTIONS_CUSTOM = ("scrolltype", "usemousewheel")
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = (
        "class_",
        "scrolltype",
    )

    def get_child_master(self):
        return self.widget.innerframe

    def configure(self, target=None):
        super().configure(self.widget.innerframe)

    def _container_layout(self, target, container_manager, properties):
        super()._container_layout(
            self.widget.innerframe, container_manager, properties
        )

    def _set_property(self, target_widget, pname, value):
        if pname in ("usemousewheel",):
            super(TTKScrolledFrameBO, self)._set_property(
                self.widget, pname, value
            )
        else:
            super(TTKScrolledFrameBO, self)._set_property(
                target_widget, pname, value
            )

    #
    # Code generation methods
    #
    def code_child_master(self):
        return "{0}.innerframe".format(self.code_identifier())

    def code_configure(self, targetid=None):
        realtarget = "{0}.innerframe".format(self.code_identifier())
        return super(TTKScrolledFrameBO, self).code_configure(realtarget)

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "usemousewheel":
            nvalue = "{0}.configure({1}={2})".format(
                self.code_identifier(), pname, tk.getboolean(value)
            )
            code_bag[pname] = [nvalue]
        else:
            super(TTKScrolledFrameBO, self)._code_set_property(
                targetid, pname, value, code_bag
            )


_builder_id = f"{_plugin_uid}.ScrolledFrame"
register_widget(
    _builder_id,
    TTKScrolledFrameBO,
    "ScrolledFrame",
    (_tab_widgets_label, "ttk"),
    group=0,
)

_builder_old = "pygubu.builder.widgets.scrolledframe"
register_widget(
    _builder_old,
    TTKScrolledFrameBO,
    public=False,
)



import tkinter as tk
from .builder import FormBuilder
from .widget import FieldWidget, WidgetInfo


EMPTY_VALUES = (None, "", [], (), {})


class TkFormBuilder(FormBuilder):
    def scan_widgets(self):
        self.search_widgets(self)

    def search_widgets(self, master=None):
        if master is None:
            master = self
        for widget in master.winfo_children():
            if isinstance(widget, FieldWidget):
                self.widgets[widget.field_name] = widget
            elif isinstance(widget, WidgetInfo):
                self.widgets_info[widget.field_name] = widget
            else:
                self.search_widgets(widget)


class WidgetViewManager:
    """A class to provide default behavior and allow user to inject
    custom code."""

    def mark_invalid(self, widget: tk.Widget, state: bool):
        ...

    def is_disabled(self, widget) -> bool:
        ...

    def show_error(self, widget: tk.Widget):
        ...

    def show_help(self, widget: tk.Widget):
        ...


class TkWidgetViewManager(WidgetViewManager):
    """Default view manager for tk widgets

    To mark widget as invalid, uses option database
    to get color information:
       form_error_bg: background color
       form_error_fg: foreground color
    If color information is not found, does nothing.
    """

    def mark_invalid(self, widget: tk.Widget, state: bool):
        class_ = widget.winfo_class()
        error_bg = widget.option_get("form_error_bg", class_)
        error_fg = widget.option_get("form_error_fg", class_)

        if not all((error_bg, error_fg)):
            # no configuration found, do nothing
            return
        if not hasattr(widget, "_oldbg"):
            widget._oldbg = None
            widget._oldfg = None
        if state:
            widget._oldbg = widget.cget("background")
            widget._oldfg = widget.cget("foreground")
            widget.configure(background=error_bg, foreground=error_fg)
        elif widget._oldbg is not None:
            widget.configure(background=widget._oldbg, foreground=widget._oldfg)

    def is_disabled(self, widget) -> bool:
        return "disabled" == widget.cget("state")


class TkWidgetBase(FieldWidget):
    view_manager = TkWidgetViewManager()

    def wis_disabled(self) -> bool:
        return self.view_manager.is_disabled(self)

    def wmark_invalid(self, state: bool):
        self.view_manager.mark_invalid(self, state)


class TkVarBasedWidget(TkWidgetBase):
    tkvar_pname = "textvariable"
    tkvar_class = tk.StringVar

    def __init__(self, *args, **kw):
        user_var = kw.get(self.tkvar_pname, None)
        if user_var is None:
            self._data_var = self.tkvar_class()
            kw[self.tkvar_pname] = self._data_var
        #
        # Some widgets can use diferent variable types such as checkbutton
        elif isinstance(user_var, tk.Variable):
            self._data_var = user_var
            self.tkvar_class = type(user_var)
        else:
            raise ValueError("Incorrect type for data variable")

        super().__init__(*args, **kw)

    def configure(self, cnf=None, **kw):
        if self.tkvar_pname in kw:
            self._data_var = kw[self.tkvar_pname]
        return super().configure(cnf, **kw)

    def wset_value(self, value):
        if value in EMPTY_VALUES:
            value = ""
        # FIXME avoid TclErrors when using typed variables Int, Double, Boolean
        tk.Variable.set(self._data_var, value)

    def wget_value(self):
        # FIXME avoid TclErrors when using typed variables Int, Double, Boolean
        return tk.Variable.get(self._data_var)


class Text(TkWidgetBase, tk.Text):
    def wset_value(self, value):
        self.delete("0.0", tk.END)
        state = self.cget("state")
        if state == tk.DISABLED:
            self.configure(state=tk.NORMAL)
            self.insert("0.0", value)
            self.configure(state=tk.DISABLED)
        else:
            self.insert("0.0", value)

    def wget_value(self):
        return self.get("0.0", "end-1c")



import tkinter as tk
from pygubu.api.v1 import IDesignerPlugin
from pygubu.utils.widget import crop_widget

from pygubu.stockimage import StockRegistry, StockImageCache, StockImage
import pygubu.plugins.tkmt.designer.preview as preview
import pygubu.plugins.tkmt.designer.properties
from .toplevelpreview import ThemedTKinterFrameTLPreview


class PygubuDesignerPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid.startswith("tkmt."):
            notused, class_name = builder_uid.split(".")
            preview_name = f"{class_name}PreviewBO"
            if hasattr(preview, preview_name):
                return getattr(preview, preview_name)
            # else:
            #    print(f"Warning: {preview_name} NOT defined")
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        top = None
        toplevel_uids = ("tkmt.ThemedTKinterFrame",)
        if builder_uid in toplevel_uids:
            ui = builder.uidefinition
            xpath = f".//object[@class='{builder_uid}']"
            node = ui.root.find(xpath)
            if node is not None:
                node.set("class", "tkmt.ThemedTKinterFrameTLPreview")

            # for a new tk root, create a diferent image cache:
            def on_root_created(root):
                image_cache = StockImageCache(root, StockImage.registry)
                builder.image_cache = image_cache

            builder.on_first_object = on_root_created
            top = builder.get_object(widget_id)

        return top

    #    def configure_for_preview(self, builder_uid: str, widget):
    #        """Make widget just display with minimal functionality."""

    def ensure_visibility_in_preview(self, builder, selected_uid: str):
        """Ensure visibility of selected_uid in preview."""
        xpath = ".//object[@class='tkmt.NotebookTab']"
        # find all tabs
        tabs = builder.uidefinition.root.findall(xpath)
        if tabs is None:
            return

        for tab in tabs:
            tab_id = tab.get("id")
            activate_tab = False
            # Check if this tab was clicked
            if tab_id == selected_uid:
                activate_tab = True
            else:
                # check if selected_uid is inside this tab
                xpath = f".//object[@id='{selected_uid}']"
                o = tab.find(xpath)
                if o is not None:
                    activate_tab = True
            if activate_tab:
                tab_builder = builder.objects[tab_id]
                tab_widget = tab_builder.widget
                notebook = tab_widget.nametowidget(tab_widget.winfo_parent())
                notebook.select(tab_widget)
                notebook.update()
                break



from pygubu.api.v1 import BuilderLoaderPlugin


class StandardTKWidgetsLoader(BuilderLoaderPlugin):
    module_default = "pygubu.plugins.tk.tkstdwidgets"
    module_map = {
        module_default: tuple(),
        "pygubu.plugins.tk.scrolledtext_bo": (
            "tk.ScrolledText",
            "pygubu.builder.widgets.tkinterscrolledtext",  # old name
        ),
    }

    def do_activate(self) -> bool:
        return True

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return self.module_default

    def can_load(self, identifier: str) -> bool:
        for module, builders in self.module_map.items():
            if identifier in builders:
                return True
        return identifier.startswith("tk.")



import os
import json
import tkinter as tk
import TKinterModernThemes as tkmt
import TKinterModernThemes.WidgetFrame as tkmt_widgets

from pygubu.i18n import _
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.utils.datatrans import ListDTO
from ..tkmt import _designer_tab_label, _plugin_uid
from .base import (
    TkmtWidgetBO,
    WidgetAsMethodBO,
    CommandProxy,
    GROUP_ROOT,
    GROUP_CONTAINER,
    GROUP_DISPLAY,
    GROUP_INPUT,
)


running_in_designer = os.getenv("PYGUBU_DESIGNER_RUNNING")


class ThemedTKinterFrameBO(TkmtWidgetBO):
    allow_bindings = False
    layout_required = False
    allowed_parents = ("root",)
    class_ = tkmt.ThemedTKinterFrame
    container = True
    pos_args = ("title",)
    kw_args = ("theme", "mode", "usecommandlineargs", "useconfigfile")
    properties = pos_args + kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"title": self.wmeta.identifier}


_builder_uid = f"{_plugin_uid}.ThemedTKinterFrame"
_themedtkinterframe = _builder_uid
register_widget(
    _builder_uid,
    ThemedTKinterFrameBO,
    "ThemedTKinterFrame",
    ("ttk", _designer_tab_label),
    GROUP_ROOT,
)


class FrameBO(WidgetAsMethodBO):
    container = True
    master_add_method = "addFrame"
    pos_args = ("name",)
    properties = pos_args + WidgetAsMethodBO.kw_args

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"name": self.wmeta.identifier}


_builder_uid = f"{_plugin_uid}.Frame"
_frame = _builder_uid
register_widget(
    _builder_uid,
    FrameBO,
    "Frame",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)

FrameBO.add_allowed_parent(_themedtkinterframe)
FrameBO.add_allowed_parent(_frame)


class LabelFrameBO(WidgetAsMethodBO):
    container = True
    master_add_method = "addLabelFrame"
    pos_args = ("text",)
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"text": self.wmeta.identifier}


_builder_uid = f"{_plugin_uid}.LabelFrame"
_labelframe = _builder_uid
register_widget(
    _builder_uid,
    LabelFrameBO,
    "LabelFrame",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)

FrameBO.add_allowed_parent(_themedtkinterframe)
FrameBO.add_allowed_parent(_frame)
FrameBO.add_allowed_parent(_labelframe)


class FrameNextColBO(BuilderObject):
    allow_bindings = False
    layout_required = False

    def realize(self, parent, extra_init_args: dict = None):
        self.widget = parent.widget
        master = parent.get_child_master()
        master.nextCol()
        return self.widget

    def code_realize(self, boparent, code_identifier=None):
        lines = []
        master = boparent.code_child_master()
        s = f"{master}.nextCol()"
        lines.append(s)
        return lines


_builder_uid = f"{_plugin_uid}.FrameNextCol"
register_widget(
    _builder_uid,
    FrameNextColBO,
    "Frame.NextCol",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)

FrameNextColBO.add_allowed_parent(_themedtkinterframe)
FrameNextColBO.add_allowed_parent(_frame)
FrameNextColBO.add_allowed_parent(_labelframe)


class SeparatorBO(WidgetAsMethodBO):
    master_add_method = "Seperator"


_builder_uid = f"{_plugin_uid}.Separator"
register_widget(
    _builder_uid,
    SeparatorBO,
    "Separator",
    ("ttk", _designer_tab_label),
    GROUP_DISPLAY,
)


class ButtonBO(WidgetAsMethodBO):
    master_add_method = "Button"
    pos_args = ("text", "command")
    kw_args = ("args",) + WidgetAsMethodBO.kw_args
    properties = pos_args + kw_args
    ro_properties = properties
    command_properties = ("command",)

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "text": self.wmeta.identifier,
            "command": None,
        }


_builder_uid = f"{_plugin_uid}.Button"
register_widget(
    _builder_uid, ButtonBO, "Button", ("ttk", _designer_tab_label), GROUP_INPUT
)


class AccentButtonBO(ButtonBO):
    master_add_method = "AccentButton"


_builder_uid = f"{_plugin_uid}.AccentButton"
register_widget(
    _builder_uid,
    AccentButtonBO,
    "AccentButton",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class CheckbuttonBO(WidgetAsMethodBO):
    master_add_method = "Checkbutton"
    pos_args = ("text", "variable")
    kw_args = ("command", "args", "disabled") + WidgetAsMethodBO.kw_args
    properties = pos_args + kw_args
    ro_properties = properties
    command_properties = ("command",)

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "text": self.wmeta.identifier,
            "variable": None,
        }


_builder_uid = f"{_plugin_uid}.Checkbutton"
register_widget(
    _builder_uid,
    CheckbuttonBO,
    "Checkbutton",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class ToggleButtonBO(CheckbuttonBO):
    master_add_method = "ToggleButton"


_builder_uid = f"{_plugin_uid}.ToggleButton"
register_widget(
    _builder_uid,
    ToggleButtonBO,
    "ToggleButton",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class SlideSwitchBO(CheckbuttonBO):
    master_add_method = "SlideSwitch"


_builder_uid = f"{_plugin_uid}.SlideSwitch"
register_widget(
    _builder_uid,
    SlideSwitchBO,
    "SlideSwitch",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class RadiobuttonBO(CheckbuttonBO):
    master_add_method = "Radiobutton"
    pos_args = ("text", "variable", "value")
    properties = pos_args + CheckbuttonBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "text": self.wmeta.identifier,
            "variable": None,
            "value": 0,
        }


_builder_uid = f"{_plugin_uid}.Radiobutton"
register_widget(
    _builder_uid,
    RadiobuttonBO,
    "Radiobutton",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class EntryBO(WidgetAsMethodBO):
    master_add_method = "Entry"
    pos_args = ("textvariable",)
    kw_args = (
        "validatecommand",
        "validatecommandargs",
        "validatecommandmode",
        "invalidcommand",
        "invalidcommandargs",
        "validate",
    ) + WidgetAsMethodBO.kw_args
    properties = pos_args + kw_args
    ro_properties = properties
    command_properties = ("validatecommand", "invalidcommand")

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "textvariable": None,
        }


_builder_uid = f"{_plugin_uid}.Entry"
register_widget(
    _builder_uid, EntryBO, "Entry", ("ttk", _designer_tab_label), GROUP_INPUT
)


class NumericalSpinboxBO(WidgetAsMethodBO):
    master_add_method = "NumericalSpinbox"
    pos_args = ("lower", "upper", "increment", "variable")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "lower": 0,
            "upper": 10,
            "increment": 1,
            "variable": None,
        }

    def _process_property_value(self, pname, value):
        if pname in ("lower", "upper", "increment"):
            return float(value)
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.NumericalSpinbox"
register_widget(
    _builder_uid,
    NumericalSpinboxBO,
    "NumericalSpinbox",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class NonnumericalSpinboxBO(WidgetAsMethodBO):
    master_add_method = "NonnumericalSpinbox"
    pos_args = ("values", "variable")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties
    json_to_list = ListDTO([], ["values should be a json list"])

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "values": [],
            "variable": None,
        }

    def _process_property_value(self, pname, value):
        if pname == "values":
            return self.json_to_list.transform(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return str(self.json_to_list.transform(value))
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.NonnumericalSpinbox"
register_widget(
    _builder_uid,
    NonnumericalSpinboxBO,
    "NonnumericalSpinbox",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class TreeviewBO(WidgetAsMethodBO):
    master_add_method = "Treeview"
    pos_args = ("columnnames", "columnwidths", "height", "data", "subentryname")
    kw_args = (
        "datacolumnnames",
        "openkey",
        "anchor",
        "newframe",
    ) + WidgetAsMethodBO.kw_args
    properties = pos_args + kw_args
    ro_properties = properties
    json_to_colname = ListDTO([], ["values should be a json list"])
    json_to_colwidth = ListDTO([], [])

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "columnnames": [],
            "columnwidths": [],
            "height": 5,
            "data": {},
            "subentryname": "",
        }

    def _process_property_value(self, pname, value):
        if pname in ("columnnames", "datacolumnnames"):
            return self.json_to_colname.transform(value)
        if pname == "columnwidths":
            return self.json_to_colwidth.transform(value)
        if pname == "newframe":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in ("columnnames", "datacolumnnames"):
            return str(self.json_to_colname.transform(value))
        if pname == "columnwidths":
            return str(self.json_to_colwidth.transform(value))
        return super()._code_process_property_value(targetid, pname, value)

    def _post_process_properties(self, tkmaster: tk.Widget, pbag: dict) -> None:
        if running_in_designer:
            self._fix_initargs_in_designer(pbag)

    def _fix_initargs_in_designer(self, kargs: dict) -> None:
        key_names = "columnnames"
        key_widths = "columnwidths"
        key_datacn = "datacolumnnames"

        # Fix names and widths
        if key_names in kargs and key_widths in kargs:
            name_count = len(kargs[key_names])
            width_count = len(kargs[key_widths])

            if width_count > name_count:
                kargs[key_widths] = kargs[key_widths][:name_count]
            if width_count < name_count:
                diff = name_count - width_count
                for i in range(0, diff):
                    kargs[key_widths].append(100)
        elif key_names in kargs and key_widths not in kargs:
            name_count = len(kargs[key_names])
            kargs[key_widths] = []
            for i in range(0, name_count):
                kargs[key_widths].append(100)
        elif key_names not in kargs and key_widths in kargs:
            width_count = len(kargs[key_widths])
            kargs[key_names] = []
            for i in range(0, width_count):
                kargs[key_names].append(f"column {i+1}")

        # fix datacolumnnames if we are in designer editing.
        if key_datacn in kargs:
            name_count = len(kargs[key_names])
            dcn_count = len(kargs[key_datacn])
            if name_count != dcn_count:
                # just clear data names until user enter correct ones
                kargs[key_datacn] = None

        # Fix data if we are in designer and no resource is set.
        key_data = "data"
        if key_data in kargs:
            kargs[key_data] = {}


_builder_uid = f"{_plugin_uid}.Treeview"
register_widget(
    _builder_uid,
    TreeviewBO,
    "Treeview",
    ("ttk", _designer_tab_label),
    GROUP_DISPLAY,
)


class OptionMenuBO(WidgetAsMethodBO):
    master_add_method = "OptionMenu"
    pos_args = ("values", "variable")
    kw_args = ("command", "args") + WidgetAsMethodBO.kw_args
    properties = pos_args + kw_args
    ro_properties = properties
    command_properties = ("command",)
    jlist_values = ListDTO([], ["values should be a json list"])

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "values": ["Test Item"],
            "variable": tk.StringVar(master),
        }

    def _process_property_value(self, pname, value):
        if pname == "values":
            return self.jlist_values.transform(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return str(self.jlist_values.transform(value))
        return super()._code_process_property_value(targetid, pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        # arg for command callback
        fargs = ["value"]
        user_args = self.wmeta.properties.get("args", None)
        if user_args:
            user_args = self._process_property_value("args", user_args)
        if user_args:
            for idx, arg in enumerate(user_args):
                fargs.insert(0, f"arg{idx}")
        return fargs


_builder_uid = f"{_plugin_uid}.OptionMenu"
register_widget(
    _builder_uid,
    OptionMenuBO,
    "OptionMenu",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class ComboboxBO(WidgetAsMethodBO):
    master_add_method = "Combobox"
    pos_args = ("values", "variable")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties
    jlist_values = ListDTO([], ["values should be a json list"])

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "values": ["Test Item"],
            "variable": tk.StringVar(master),
        }

    def _process_property_value(self, pname, value):
        if pname == "values":
            return self.jlist_values.transform(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return str(self.jlist_values.transform(value))
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.Combobox"
register_widget(
    _builder_uid,
    ComboboxBO,
    "Combobox",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class MenuButtonBO(WidgetAsMethodBO):
    master_add_method = "MenuButton"
    pos_args = ("menu", "defaulttext")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "menu": None,
            "defaulttext": "MenuButton",
        }


_builder_uid = f"{_plugin_uid}.MenuButton"
register_widget(
    _builder_uid,
    MenuButtonBO,
    "MenuButton",
    ("ttk", _designer_tab_label),
    GROUP_INPUT,
)


class NotebookBO(WidgetAsMethodBO):
    master_add_method = "Notebook"
    pos_args = ("name",)
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties
    container = True

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "name": self.wmeta.identifier,
        }


_builder_uid = f"{_plugin_uid}.Notebook"
_notebook = _builder_uid
register_widget(
    _builder_uid,
    NotebookBO,
    "Notebook",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)


class NotebookTabBO(WidgetAsMethodBO):
    master_add_method = "addTab"
    pos_args = ("text",)
    kw_args = ()
    properties = pos_args + ("makeResizable",)
    ro_properties = pos_args
    container = True

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "text": self.wmeta.identifier,
        }


_builder_uid = f"{_plugin_uid}.NotebookTab"
_notebok_tab = _builder_uid
register_widget(
    _builder_uid,
    NotebookTabBO,
    "Notebook.Tab",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)

NotebookBO.add_allowed_child(_notebok_tab)
NotebookTabBO.add_allowed_parent(_notebook)


class PanedWindowBO(WidgetAsMethodBO):
    master_add_method = "PanedWindow"
    pos_args = ("name", "orient")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties
    container = True

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "name": self.wmeta.identifier,
        }


_builder_uid = f"{_plugin_uid}.PanedWindow"
_panedwindow = _builder_uid
register_widget(
    _builder_uid,
    PanedWindowBO,
    "PanedWindow",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)


class PanedWindowPaneBO(WidgetAsMethodBO):
    master_add_method = "addWindow"
    pos_args = ("weight",)
    properties = pos_args
    ro_properties = properties
    container = True

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "weight": 1,
        }


_builder_uid = f"{_plugin_uid}.PanedWindowPane"
_panedwindow_pane = _builder_uid
register_widget(
    _builder_uid,
    PanedWindowPaneBO,
    "PanedWindow.Pane",
    ("ttk", _designer_tab_label),
    GROUP_CONTAINER,
)

PanedWindowBO.add_allowed_child(_panedwindow_pane)
PanedWindowPaneBO.add_allowed_parent(_panedwindow)


class BlankBO(WidgetAsMethodBO):
    master_add_method = "Blank"
    pos_args = ("name",)
    kw_args = ("row", "col", "rowspan", "colspan")
    properties = pos_args + kw_args
    ro_properties = properties

    def realize(self, parent, extra_init_args: dict = None):
        super().realize(parent, extra_init_args)
        self.widget = parent.widget
        return self.widget

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "name": "Blank",
        }


_builder_uid = f"{_plugin_uid}.Blank"
register_widget(
    _builder_uid, BlankBO, "Blank", ("ttk", _designer_tab_label), GROUP_DISPLAY
)


class LabelBO(WidgetAsMethodBO):
    master_add_method = "Label"
    pos_args = ("text",)
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"text": self.wmeta.identifier}


_builder_uid = f"{_plugin_uid}.Label"
_labelframe = _builder_uid
register_widget(
    _builder_uid, LabelBO, "Label", ("ttk", _designer_tab_label), GROUP_DISPLAY
)


class TextBO(WidgetAsMethodBO):
    master_add_method = "Text"
    pos_args = ("text",)
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {"text": self.wmeta.identifier}


_builder_uid = f"{_plugin_uid}.Text"
_labelframe = _builder_uid
register_widget(
    _builder_uid, TextBO, "Text", ("ttk", _designer_tab_label), GROUP_DISPLAY
)


class ScaleBO(WidgetAsMethodBO):
    master_add_method = "Scale"
    pos_args = ("lower", "upper", "variable")
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "lower": 0,
            "upper": 10,
            "variable": None,
        }

    def _process_property_value(self, pname, value):
        if pname in ("lower", "upper"):
            return float(value)
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.Scale"
register_widget(
    _builder_uid, ScaleBO, "Scale", ("ttk", _designer_tab_label), GROUP_INPUT
)


class ProgressbarBO(WidgetAsMethodBO):
    master_add_method = "Progressbar"
    pos_args = ("variable",)
    properties = pos_args + WidgetAsMethodBO.kw_args
    ro_properties = properties

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {
            "variable": None,
        }

    def _process_property_value(self, pname, value):
        if pname in ("lower", "upper"):
            return float(value)
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.Progressbar"
register_widget(
    _builder_uid,
    ProgressbarBO,
    "Progressbar",
    ("ttk", _designer_tab_label),
    GROUP_DISPLAY,
)

# TODO: matplotlibFrame



import tkinter.ttk as ttk
from pygubu.utils.widget import HideableMixin
from .builder import FormBuilder
from .widget import FieldWidget, WidgetInfo, FormInfo
from .tkwidget import TkFormBuilder, TkVarBasedWidget, TkWidgetViewManager


class FrameFormBuilder(TkFormBuilder, ttk.Frame):
    ...


class TtkWidgetViewManager(TkWidgetViewManager):
    """Default view manager for ttk widgets

    To mark widget as invalid, combines the word "Error."
    with the widget style class.
    """

    def mark_invalid(self, widget: ttk.Widget, state: bool):
        style: str = widget.cget("style")
        if style:
            # remove prefix if exists.
            prefix = "Error."
            if style.startswith(prefix):
                style = style[len(prefix) :]
            if state:
                style = f"{prefix}{style}"
            widget.configure(style=style)


class TtkVarBasedWidget(TkVarBasedWidget):
    view_manager = TtkWidgetViewManager()


class TtkWidgetInfoViewManager:
    def _set_style(self, widget, mode: str):
        style: str = widget.cget("style")
        if style:
            # remove prefix if exists
            prefix = "Error."
            if style.startswith(prefix):
                style = style[len(prefix) :]
            # remove prefix if exists
            prefix = "Help."
            if style.startswith(prefix):
                style = style[len(prefix) :]
            style = f"{mode}.{style}"
            widget.configure(style=style)

    def show_error(self, widget: ttk.Widget):
        self._set_style(widget, "Error")

    def show_help(self, widget: ttk.Widget):
        self._set_style(widget, "Help")


class LabelWidgetInfo(WidgetInfo, HideableMixin, ttk.Label):
    """Used to display help and errors messages for the associated form field."""

    view_manager = TtkWidgetInfoViewManager()

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

    def show_error(self, error):
        self.hidden = False
        self.configure(text=error.message)
        self.view_manager.show_error(self)

    def show_help(self, message):
        self.hidden = False
        self.configure(text=message)
        self.view_manager.show_help(self)

    def clear(self):
        self.configure(text="")
        self.hidden = True


class Entry(TtkVarBasedWidget, ttk.Entry):
    pass


class Label(TtkVarBasedWidget, ttk.Label):
    """A Display only field using ttk.Label"""

    def __init__(self, *args, **kw):
        self.original_data = None
        super().__init__(*args, **kw)

    def wget_value(self):
        return self.original_data

    def wset_value(self, value):
        self.original_data = value
        super().wset_value(str(value))


class Checkbutton(TtkVarBasedWidget, ttk.Checkbutton):
    tkvar_pname = "variable"


class Combobox(TtkVarBasedWidget, ttk.Combobox):
    pass
